from __future__ import annotations

from typing import Optional, cast

from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.positioned_object import PositionedObject, render_cached_property
from neoscore.core.units import ZERO, Unit
from neoscore.western import notehead_tables
from neoscore.western.duration import Duration, DurationDef
from neoscore.western.duration_display import DurationDisplay
from neoscore.western.notehead_tables import NoteheadTable
from neoscore.western.pitch import Pitch, PitchDef
from neoscore.western.staff_object import StaffObject


class Notehead(MusicText, StaffObject):

    """A simple notehead automatically selected and vertically positioned."""

    def __init__(
        self,
        pos_x: Unit,
        parent: PositionedObject,
        pitch: PitchDef,
        duration: DurationDef,
        font: Optional[MusicFont] = None,
        table: NoteheadTable = notehead_tables.STANDARD,
        glyph_override: Optional[str] = None,
    ):
        """
        Args:
            pos_x: The x-axis position relative to ``parent``.
                The y-axis position is calculated automatically based
                on ``pitch`` and contextual information in ``self.staff``.
            parent: Must be a :obj:`.Staff` or a descendant of one.
            pitch: The pitch used to vertically place the notehead.
                See :obj:`.Pitch` and :obj:`.PitchDef`
            duration: The logical duration of the notehead. This is used to determine
                the glyph style.
            font: If provided, this overrides any font found in the ancestor chain.
            table: The set of noteheads to use according to ``duration``.
                See :obj:`.notehead_tables`.
            glyph_override: A SMuFL glyph name. If given, this overrides
                the glyph normally looked up with ``duration`` from ``table``.
        """
        self.pitch = pitch
        self.duration = duration
        self._table = table
        self._glyph_override = glyph_override
        MusicText.__init__(
            self,
            (pos_x, ZERO),
            parent,
            "",
            font,
        )
        StaffObject.__init__(self, parent)
        self._update_music_text()

    @property
    def visual_width(self) -> Unit:
        return self.bounding_rect.width

    @property
    def pitch(self) -> Pitch:
        return self._pitch

    @pitch.setter
    def pitch(self, value: PitchDef):
        rebuild_needed = hasattr(self, "_pitch")
        self._pitch = Pitch.from_def(value)
        if rebuild_needed:
            self._update_music_text()

    @property
    def duration(self) -> Duration:
        return self._duration

    @duration.setter
    def duration(self, value: DurationDef):
        rebuild_needed = hasattr(self, "_duration")
        value = Duration.from_def(value)
        if value.display is None:
            raise ValueError(f"{value} cannot be represented as a single note")
        self._duration = value
        if rebuild_needed:
            self._update_music_text()

    @property
    def table(self) -> NoteheadTable:
        return self._table

    @table.setter
    def table(self, value: NoteheadTable):
        self._table = value
        self._update_music_text()

    @property
    def glyph_override(self) -> Optional[str]:
        return self._glyph_override

    @glyph_override.setter
    def glyph_override(self, value: Optional[str]):
        self._glyph_override = value
        self._update_music_text()

    @render_cached_property
    def staff_pos(self) -> Unit:
        """The y-axis position in the staff.

        This is derived from the ``pitch`` and the staff's active clef at this object's
        position.
        """
        return self.staff.middle_c_at(self.pos_x_in_staff) + self.staff.unit(
            self.pitch.staff_pos_from_middle_c
        )

    def _update_music_text(self):
        duration_display = cast(DurationDisplay, self.duration.display)
        if self._glyph_override:
            glyph = self._glyph_override
        else:
            glyph = self._table.lookup_duration(duration_display.base_duration)
        self.text = glyph
        self.y = self.staff_pos - self.staff.map_to(self.parent).y



from dataclasses import dataclass
from typing import List, Optional

from neoscore.core.color import ColorDef
from neoscore.core.pen_pattern import PenPattern
from neoscore.core.units import Union, Unit


@dataclass(frozen=True)
class BarlineStyle:
    """Style for an individual sub-barline.

    Use multiple of these to specify multipart lines like double barlines.
    """

    thickness: Union[str, float, Unit] = 0.16
    """Thickness for this line

    This can be defined either as a SMuFL engraving default key, a float value of
    pseudo-staff-units, or a plain unit value.
    """

    gap_right: Union[str, float, Unit] = 0.4
    """The gap to the right of this line, if another follows.

    Like ``thickness``, this can be defined either as a SMuFL engraving default key, a
    float value of pseudo-staff-units, or a plain unit value.
    """

    pattern: PenPattern = PenPattern.SOLID
    """Line pattern of this line"""

    color: Optional[ColorDef] = None
    """Color of this line given as a :obj:`.ColorDef`"""


# NOTE: When adding a new style, be sure to add it to ALL_STYLES below

SINGLE = [BarlineStyle("thinBarlineThickness")]
"""A plain single barline."""

DASH = [BarlineStyle("thinBarlineThickness", pattern=PenPattern.DASH)]
"""A single dashed barline."""

DOT = [BarlineStyle("thinBarlineThickness", pattern=PenPattern.DOT)]
"""A single dotted barline."""

THIN_DOUBLE = [
    BarlineStyle("thinBarlineThickness", "barlineSeparation"),
    BarlineStyle("thinBarlineThickness"),
]
"""A thin double barline."""

END = [
    BarlineStyle("thinBarlineThickness", "thickBarlineSeparation"),
    BarlineStyle("thickBarlineThickness"),
]
"""A thick double barline conventionally used at the end of scores."""

ALL_STYLES: List[List[BarlineStyle]] = [SINGLE, DASH, DOT, THIN_DOUBLE, END]



from dataclasses import dataclass, field

from PyQt5 import QtGui

from neoscore.core.color import Color
from neoscore.core.pen_cap_style import PenCapStyle
from neoscore.core.pen_join_style import PenJoinStyle
from neoscore.core.pen_pattern import PenPattern
from neoscore.core.units import Unit


@dataclass(frozen=True)
class PenInterface:
    """Interface for a pen controlling path outline appearance."""

    color: Color

    thickness: Unit
    """The pen thickness. If zero, a cosmetic pixel thickness is used."""

    pattern: PenPattern

    join_style: PenJoinStyle

    cap_style: PenCapStyle

    qt_object: QtGui.QPen = field(init=False)

    def __post_init__(self):
        color = QtGui.QColor(
            self.color.red, self.color.green, self.color.blue, self.color.alpha
        )
        brush = QtGui.QBrush(color)
        q_pen = QtGui.QPen(
            brush,
            self.thickness.base_value,
            self.pattern.value,
            self.cap_style.value,
            self.join_style.value,
        )
        super().__setattr__("qt_object", q_pen)



from __future__ import annotations

from typing import Optional, cast

from neoscore.core import neoscore
from neoscore.core.brush import Brush
from neoscore.core.directions import DirectionY
from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.layout_controllers import NewLine
from neoscore.core.music_char import MusicChar
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.path import Path
from neoscore.core.pen import Pen
from neoscore.core.pen_pattern import PenPattern
from neoscore.core.point import ORIGIN, Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.spanner import Spanner
from neoscore.core.units import Unit

_GLYPHS = {
    "15ma": "quindicesimaAlta",
    "8va": "ottavaAlta",
    "8vb": "ottavaBassaVb",
    "15mb": "quindicesimaBassaMb",
    "(": "octaveParensLeft",
    ")": "octaveParensRight",
}


class OctaveLine(Spanner, PositionedObject, HasMusicFont):

    """An octave indication with a dashed line.

    This octave line is purely cosmetic, and does not result in any automatic
    transpositions.

    At the starting position the octave is written in text, followed by a dashed line
    ending in a small vertical hook pointing toward the staff. If the spanner goes
    across line breaks, the octave text is repeated in parentheses at the line
    beginning.
    """

    def __init__(
        self,
        start: PointDef,
        start_parent: PositionedObject,
        end_x: Unit,
        end_parent: Optional[PositionedObject] = None,
        indication: str = "8va",
        direction: DirectionY = DirectionY.DOWN,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            start: The starting point.
            start_parent: The parent for the starting position. If no font is given,
                this or one of its ancestors must implement :obj:`.HasMusicFont`.
            end_x: The spanner end x position. The y position will be
                automatically calculated to be horizontal.
            end_parent: An object either in a Staff or
                a staff itself. The root staff of this *must* be the same
                as the root staff of ``start_parent``. If omitted, the
                stop point is relative to the start point.
            indication: A valid octave indication. currently supported indications are:

                    * '15ma' (two octaves higher)
                    * '8va' (one octave higher)
                    * '8vb' (one octave lower)
                    * '15mb' (two octaves lower)

            direction: The direction the line's ending hook points.
                For lines above staves, this should be down, and vice versa for below.
            font: If provided, this overrides any font found in the ancestor chain.
        """
        PositionedObject.__init__(self, start, start_parent)
        Spanner.__init__(self, end_x, end_parent or self)
        if font is None:
            font = HasMusicFont.find_music_font(start_parent)
        self._music_font = font
        self.direction = direction

        self.line_path = Path(
            ORIGIN,
            self,
            Brush.no_brush(),
            Pen(
                thickness=font.engraving_defaults["octaveLineThickness"],
                pattern=PenPattern.DASH,
            ),
        )

        self.line_text = _OctaveLineText(
            ORIGIN, self, self.breakable_length, indication, font
        )

        # Vertically center the path relative to the text
        text_rect = self.line_text.bounding_rect
        path_x = text_rect.width
        path_y = cast(Unit, text_rect.height / -2)
        self.line_path.pos = Point(path_x, path_y)

        # Drawn main line part
        self.line_path.line_to(self.end_pos.x, path_y, self.end_parent)
        self.line_path.line_to(
            self.end_pos.x,
            (path_y + font.unit(0.75 * self.direction.value)),
            self.end_parent,
        )

    @property
    def music_font(self) -> MusicFont:
        return self._music_font


class _OctaveLineText(MusicText):
    """An octave text mark recurring at line beginnings with added parenthesis.

    This is a private class meant to be used exclusively in the context
    of an OctaveLine
    """

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        length: Unit,
        indication: str,
        font: MusicFont,
    ):
        super().__init__(
            pos,
            parent,
            _GLYPHS[indication],
            font,
            background_brush=neoscore.background_brush,
        )
        open_paren_char = MusicChar(self.music_font, _GLYPHS["("])
        close_paren_char = MusicChar(self.music_font, _GLYPHS[")"])
        self.parenthesized_text = (
            open_paren_char.codepoint + self.text + close_paren_char.codepoint
        )
        self._length = length

    @property
    def breakable_length(self) -> Unit:
        return self._length

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        super().render_complete(pos, flowable_line)

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        super().render_complete(pos, flowable_line)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        super().render_complete(pos, flowable_line)



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from typing import List, Tuple, Union, cast

from typing_extensions import TypeAlias

_glyph_names = {
    0: "timeSig0",
    1: "timeSig1",
    2: "timeSig2",
    3: "timeSig3",
    4: "timeSig4",
    5: "timeSig5",
    6: "timeSig6",
    7: "timeSig7",
    8: "timeSig8",
    9: "timeSig9",
}


@dataclass(frozen=True)
class Meter:
    """A time signature graphical meter comprising two lists of glyphs.

    Users are recommended to use the ``Meter.numeric()`` constructor,
    which provides a convenient way to create standard meters. The
    direct constructor is mostly useful for writing more exotic
    meters, supporting arbitrary glyphs.

    If only a single line of text is needed for the meter (as with the
    common time "C"), it should go in ``upper_text_glyph_names`` and
    ``lower_text_glyph_names`` should be left empty.

    """

    upper_text_glyph_names: List[str]
    lower_text_glyph_names: List[str]

    @classmethod
    def numeric(cls, upper: Union[int | List[int]], lower: int) -> Meter:
        """Create a meter with upper and lower numbers.

        The upper number can be a single number, or a list of them.

        Lists of numbers will be treated as additive time signatures
        where each upper number is joined by a plus sign. This is
        useful for time signatures like ``[3 + 2 + 3] / 8``.
        """
        if isinstance(upper, int):
            upper_glyphs = Meter._glyphs_for_number(upper)
        else:
            upper_glyphs = reduce(
                lambda a, b: a + ["timeSigPlus"] + b,
                [Meter._glyphs_for_number(number) for number in upper],
            )
        lower_glyphs = Meter._glyphs_for_number(lower)
        return Meter(upper_glyphs, lower_glyphs)

    @classmethod
    def from_def(cls, meter_def: MeterDef) -> Meter:
        if isinstance(meter_def, tuple):
            return Meter.numeric(*meter_def)
        return cast(Meter, meter_def)

    @staticmethod
    def _glyphs_for_number(number: int) -> List[str]:
        return [_glyph_names[int(digit)] for digit in str(number)]


COMMON_TIME = Meter(["timeSigCommon"], [])
CUT_TIME = Meter(["timeSigCutCommon"], [])


MeterDef: TypeAlias = Union[Meter, Tuple[Union[int, List[int]], int]]
"""Shorthand for Meter.

Either a Meter, or an argument tuple for ``Meter.numeric``.
"""



from dataclasses import dataclass, field

from PyQt5 import QtGui

from neoscore.core.brush_pattern import BrushPattern
from neoscore.core.color import Color


@dataclass(frozen=True)
class BrushInterface:
    """Interface for a generic drawing brush controlling fill patterns."""

    color: Color
    pattern: BrushPattern
    qt_object: QtGui.QBrush = field(init=False)

    def __post_init__(self):
        color = QtGui.QColor(
            self.color.red, self.color.green, self.color.blue, self.color.alpha
        )
        q_brush = QtGui.QBrush(color, self.pattern.value)
        super().__setattr__("qt_object", q_brush)



from __future__ import annotations

from typing import Any, List, Optional, Tuple

from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.layout_controllers import NewLine
from neoscore.core.music_font import MusicFont
from neoscore.core.painted_object import PaintedObject
from neoscore.core.path import Path
from neoscore.core.pen import PenDef
from neoscore.core.point import Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO, Unit
from neoscore.western.staff_fringe_layout import StaffFringeLayout
from neoscore.western.staff_group import StaffGroup


class AbstractStaff(PaintedObject, HasMusicFont):
    """An abstract superclass for staves.

    This is not meant to be used directly."""

    # Type sentinel used to hackily check if objects are Staff
    # without importing the type, risking cyclic imports.
    _neoscore_abstract_staff_type_marker = True

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        length: Unit,
        group: Optional[StaffGroup],
        line_spacing: Unit,
        line_count: int,
        music_font: Optional[MusicFont],
        pen: Optional[PenDef],
    ):
        self._music_font = music_font
        PaintedObject.__init__(self, pos, parent, pen=pen)
        self._line_spacing = line_spacing
        self._line_count = line_count
        self._length = length
        self._group = group or StaffGroup()
        self._group.staves.add(self)

    @property
    def music_font(self) -> MusicFont:
        return self._music_font

    @property
    def height(self) -> Unit:
        """The height of the staff from top to bottom line.

        If the staff only has one line, its height is defined as 0.
        """
        return self.line_spacing * (self.line_count - 1)

    @property
    def line_spacing(self) -> Unit:
        """The distance between two lines in the staff."""
        return self._line_spacing

    @property
    def line_count(self) -> int:
        """The number of lines in the staff"""
        return self._line_count

    @property
    def center_y(self) -> Unit:
        """The position of the center staff position"""
        return self.height / 2

    @property
    def barline_extent(self) -> Tuple[Unit, Unit]:
        """The starting and stopping Y positions of barlines in this staff.

        For staves with more than 1 line, this extends from the top line to bottom
        line. For single-line staves, this extends from 1 unit above and below the
        staff.
        """
        if self.line_count == 1:
            return -self.line_spacing, self.line_spacing
        else:
            return ZERO, self.height

    @property
    def breakable_length(self) -> Unit:
        """Staves are breakable over their full length"""
        # Override expensive ``Path.length`` since the staff length here
        # is already known.
        return self._length

    @property
    def group(self) -> StaffGroup:
        """The staff group this belongs to.

        A staff group should be assigned during initalization for proper display of
        staff systems.
        """
        return self._group

    def find_ordered_descendants_with_attr(self, attr: str) -> List[Tuple[Unit, Any]]:
        """Find all descendants with an attribute, sorted with their staff x positions"""
        result = [
            (self.descendant_pos_x(obj), obj)
            for obj in self.descendants_with_attribute(attr)
        ]
        result.sort(key=lambda tup: tup[0])
        return result

    def y_inside_staff(self, pos_y: Unit) -> bool:
        """Determine if a y-axis position is inside the staff.

        This is true for any position within or on the outer lines.
        """
        return ZERO <= pos_y <= self.height

    def fringe_layout_at(self, location: Optional[NewLine]) -> StaffFringeLayout:
        return self.group.fringe_layout_at(self, location)

    def _render_slice(
        self,
        pos: Point,
        clip_start_x: Optional[Unit],
        clip_width: Optional[Unit],
        flowable_line: Optional[NewLine],
    ):
        fringe_layout = self.fringe_layout_at(flowable_line)
        if clip_width is None:
            if clip_start_x is None:
                slice_length = self.breakable_length
            else:
                slice_length = self.breakable_length - clip_start_x
        else:
            slice_length = clip_width
        inside_flowable = bool(flowable_line)
        if inside_flowable:
            segment_pos = Point(pos.x + fringe_layout.staff, pos.y)
        else:
            segment_pos = Point(fringe_layout.staff, ZERO)
        path = self._create_staff_segment_path(
            segment_pos, slice_length - fringe_layout.staff, inside_flowable
        )
        path.render()
        path.remove()

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        self._render_slice(pos, None, None, flowable_line)

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        self._render_slice(
            pos,
            ZERO,
            flowable_line.flowable_x + flowable_line.length - flowable_x,
            flowable_line,
        )

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        self._render_slice(pos, object_x, flowable_line.length, flowable_line)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        self._render_slice(pos, object_x, None, flowable_line)

    def _create_staff_segment_path(
        self, pos: Point, length: Unit, inside_flowable: bool
    ) -> Path:
        parent = None if inside_flowable else self
        path = Path(pos, parent, pen=self.pen)
        line_y = ZERO
        for i in range(self.line_count):
            path.move_to(ZERO, line_y)
            path.line_to(length, line_y)
            line_y += self.line_spacing
        return path

    def register_layout_controllers(self):
        """Register any flowable margin controllers needed by the staff.

        Staff subclasses must implement this.
        """
        raise NotImplementedError

    def fringe_layout_for_isolated_staff(
        self, location: Optional[NewLine]
    ) -> StaffFringeLayout:
        """Determine the staff fringe layout of this staff in isolation.

        This is the layout needed if this staff was alone in a staff system.
        ``StaffGroup`` uses this as a starting point in its fringe layout logic.

        Staff subclasses must implement this.
        """
        raise NotImplementedError

    def pre_render_hook(self):
        super().pre_render_hook()
        self.register_layout_controllers()



from typing import List, NamedTuple, Optional, Union

from backports.cached_property import cached_property

from neoscore.core.directions import DirectionX, DirectionY
from neoscore.core.point import ORIGIN, Point
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.rect import Rect
from neoscore.core.units import ZERO, Unit
from neoscore.western import notehead_tables
from neoscore.western.accidental import Accidental
from neoscore.western.duration import Duration, DurationDef
from neoscore.western.flag import Flag
from neoscore.western.ledger_line import LedgerLine
from neoscore.western.notehead import Notehead
from neoscore.western.notehead_tables import NoteheadTable
from neoscore.western.pitch import PitchDef
from neoscore.western.rest import Rest
from neoscore.western.rhythm_dot import RhythmDot
from neoscore.western.staff import Staff
from neoscore.western.staff_object import StaffObject
from neoscore.western.stem import Stem


class PitchAndGlyph(NamedTuple):
    """Used to define individual notes with one-off SMuFL glyphs."""

    pitch: PitchDef
    """The pitch for the notehead"""

    notehead_glyph: str
    """The SMuFL glyph name for the notehead"""


class Chordrest(PositionedObject, StaffObject):

    """A chord or a rest.

    This is a unified interface for conventionally notated
    musical notes/chords/rests. It can be given any number of pitches
    to be used as notes in the chord, or ``None`` for a rest.

    It will automatically generate and lay out:

    * :obj:`.Notehead`\ s if pitches are given
    * a :obj:`.Stem` if pitches are given and required by the given :obj:`.Duration`
    * a :obj:`.Flag` if pitches are given and required by the given ``Duration``
    * :obj:`.LedgerLine`\ s as needed (taking into consideration the given
      pitches and their location on the :obj:`.Staff`)
    * :obj:`.Accidental`\ s as needed by any given pitches
    * a :obj:`.Rest` if no pitches are given
    * :obj:`.RhythmDot`\ s if needed by the given ``Duration``

    Any accidentals given in pitches will be unconditionally drawn
    regardless of context and key signature.
    """

    # All cached properties which change on rebuilds *must* be registered here.
    _CACHED_PROPS = [
        "ledger_line_positions",
        "rhythm_dot_positions",
        "furthest_notehead",
        "highest_notehead",
        "lowest_notehead",
        "leftmost_notehead",
        "rightmost_notehead",
        "widest_notehead",
        "extra_attachment_point",
        "tremolo_attachment_point",
        "notehead_column_bounding_rect",
        "noteheads_outside_staff",
        "leftmost_notehead_outside_staff",
        "rightmost_notehead_outside_staff",
        "notehead_column_outside_staff_width",
        "stem_height",
    ]

    def __init__(
        self,
        pos_x: Unit,
        staff: Staff,
        notes: Optional[List[Union[PitchDef, PitchAndGlyph]]],
        duration: DurationDef,
        rest_y: Optional[Unit] = None,
        stem_direction: Optional[DirectionY] = None,
        beam_break_depth: Optional[int] = None,
        beam_hook_dir: Optional[DirectionX] = None,
        table: NoteheadTable = notehead_tables.STANDARD,
    ):
        """
        Args:
            pos_x: The horizontal position in the staff
            staff: The staff the object is attached to
            notes: A list of pitches and optional notehead-specific data. If ``None``
                this indicates a rest. For simple notes and chords, this can typically
                be a list of pitch string shorthands (see :obj:`.Pitch.from_str`). Pitches
                with extended accidentals can be given by passing fully constructed
                ``Pitch`` objects. Individual notehead glyphs (by default taken from the
                given ``table``) can be overridden by passing a tuple of a pitch
                and a SMuFL glyph name string.
            duration: The written duration for the object.
            rest_y: The vertical position used by rests. This defaults to the center
                of the staff.
            stem_direction: An optional stem direction override. If omitted, the
                direction is automatically calculated to point away from
                the furthest-out notehead.
            beam_break_depth: Break depth used if in a :obj:`.BeamGroup`.
            beam_hook_dir: Beamlet hook direction used in a ``BeamGroup``.
            table: The set of noteheads to use according to ``duration``.
        """
        StaffObject.__init__(self, staff)
        PositionedObject.__init__(self, Point(pos_x, ZERO), staff)
        self.duration = duration
        self._noteheads = []
        self._accidentals = []
        self._ledgers = []
        self._dots = []
        self._notes = [] if notes is None else notes
        self._stem = None
        self._flag = None
        self._rest_y = rest_y
        self._rest = None
        self._stem_direction_override = stem_direction
        self._beam_break_depth = beam_break_depth
        self._beam_hook_dir = beam_hook_dir
        self._table = table
        self._rebuild()

    @property
    def notes(self) -> List[Union[PitchDef, PitchAndGlyph]]:
        return self._notes

    @notes.setter
    def notes(self, value: Optional[List[Union[PitchDef, PitchAndGlyph]]]):
        self._notes = [] if value is None else value
        self._rebuild()

    @property
    def noteheads(self) -> List[Notehead]:
        """The noteheads contained in this Chordrest."""
        return self._noteheads

    @property
    def rest_y(self) -> Optional[Unit]:
        """The vertical position used by generated rests.

        Defaults to the staff center.
        """
        return self._rest_y

    @rest_y.setter
    def rest_y(self, value: Optional[Unit]):
        self._rest_y = value
        self._rebuild()

    @property
    def rest(self) -> Optional[Rest]:
        """A rest glyph, if no noteheads exist."""
        return self._rest

    @property
    def accidentals(self) -> List[Accidental]:
        """The accidentals contained in this Chordrest."""
        return self._accidentals

    @property
    def ledgers(self) -> List[LedgerLine]:
        """The ledger lines contained in this Chordrest.

        An empty list means none are needed.
        """
        return self._ledgers

    @property
    def dots(self) -> List[RhythmDot]:
        return self._dots

    @property
    def stem(self) -> Optional[Stem]:
        """The stem for the Chordrest."""
        return self._stem

    @property
    def flag(self) -> Optional[Flag]:
        """The flag attached to the stem."""
        return self._flag

    @property
    def beam_break_depth(self) -> Optional[int]:
        """Break depth used if in a :obj:`.BeamGroup`.

        If this Chordrest is within a beam group, this triggers a beam subdivision break
        at this point. The value indicates the number of beams to which the subdivision
        breaks. For example, in run of 16th notes a ``beam_break_depth`` of ``1`` would
        indicate a subdivision break to 1 beam at this point.
        """
        return self._beam_break_depth

    @property
    def beam_hook_dir(self) -> Optional[DirectionX]:
        """Beamlet hook direction used in a :obj:`.BeamGroup`.

        If this Chordrest is within a beam group and this position is one requiring a
        beamlet hook whose direction is ambiguous, this controls that direction.
        """
        return self._beam_hook_dir

    @property
    def table(self) -> NoteheadTable:
        return self._table

    @table.setter
    def table(self, table: NoteheadTable):
        self._table = table
        self._rebuild()

    @property
    def duration(self) -> Duration:
        """The written length of this event.

        This affects many components of the chordrest.
        """
        return self._duration

    @duration.setter
    def duration(self, value: DurationDef):
        rebuild_needed = hasattr(self, "_duration")
        value = Duration.from_def(value)
        if value.display is None:
            raise ValueError(f"{value} cannot be represented as a single note")
        self._duration = value
        if rebuild_needed:
            self._rebuild()

    @cached_property
    def ledger_line_positions(self) -> List[Unit]:
        """A set of staff positions of needed ledger lines.

        Positions are in centered staff positions.

        An empty list means no ledger lines are needed.
        """
        highest = self.highest_notehead.y
        lowest = self.lowest_notehead.y
        # This could be optimized by only doing both ledger lookups if
        # ``highest`` and ``lowest`` are on opposite sides of the staff,
        # in which case the check-then-append step can be done
        # unconditionally.
        ledgers = self.staff.ledgers_needed_for_y(lowest)
        for ledger_pos in self.staff.ledgers_needed_for_y(highest):
            if ledger_pos not in ledgers:
                ledgers.append(ledger_pos)
        return ledgers

    @cached_property
    def rhythm_dot_positions(self) -> List[Point]:
        """The positions of all rhythm dots needed."""
        start_padding = self.staff.unit(0.25)
        if self.rest:
            dot_start_x = self.rest.x + self.rest.bounding_rect.width + start_padding
            dottables = {self.rest}
        else:
            dot_start_x = (
                self.leftmost_notehead.x
                + self.notehead_column_bounding_rect.width
                + start_padding
            )
            dottables = self.noteheads
        result = []
        for i in range(self.duration.display.dot_count):
            for dottable in dottables:
                if dottable.y.display_value % 1 == 0:
                    # Dottable is on a line, add dot offset to space below
                    y_offset = self.staff.unit(-0.5)
                else:
                    y_offset = self.staff.unit(0)
                result.append(
                    Point(
                        dot_start_x + (self.staff.unit(0.5) * i), dottable.y + y_offset
                    )
                )
        return result

    @cached_property
    def furthest_notehead(self) -> Optional[Notehead]:
        """The notehead furthest from the staff center"""
        return max(
            self.noteheads,
            key=lambda n: abs(n.y - self.staff.center_y),
            default=None,
        )

    @cached_property
    def highest_notehead(self) -> Optional[Notehead]:
        """The highest notehead in the chord."""
        return min(self.noteheads, key=lambda n: n.y, default=None)

    @cached_property
    def lowest_notehead(self) -> Optional[Notehead]:
        """The lowest notehead in the chord."""
        return max(self.noteheads, key=lambda n: n.y, default=None)

    @cached_property
    def leftmost_notehead(self) -> Optional[Notehead]:
        """The notehead furthest to the left in the chord"""
        return min(self.noteheads, key=lambda n: n.x, default=None)

    @cached_property
    def rightmost_notehead(self) -> Optional[Notehead]:
        """The notehead furthest to the right in the chord"""
        return max(self.noteheads, key=lambda n: n.x, default=None)

    @cached_property
    def extra_attachment_point(self) -> Point:
        """A point where common attachments like ornaments could go.

        For chords, this is a point centered above or below the outermost notehead
        opposite of the stem direction.

        For rests, this is a point centered above the rest.

        The returned point is relative to the Chordrest.
        """
        if self.rest:
            bounding_rect = self.rest.bounding_rect
            x = self.rest.x + bounding_rect.x + (bounding_rect.width / 2)
            y = self.rest.y + bounding_rect.y - self.staff.unit(1)
            return Point(x, y)
        if self.stem_direction == DirectionY.UP:
            notehead = self.lowest_notehead
            bounding_rect = notehead.bounding_rect
            y = notehead.y + bounding_rect.y + bounding_rect.height + self.staff.unit(1)
        else:
            notehead = self.highest_notehead
            bounding_rect = notehead.bounding_rect
            y = notehead.y + bounding_rect.y - self.staff.unit(1)
        x = notehead.x + bounding_rect.x + (bounding_rect.width / 2)
        return Point(x, y)

    @cached_property
    def tremolo_attachment_point(self) -> Point:
        """A convenient reasonable point for tremolos to be placed.

        The returned point is relative to the chordrest.

        The precise position returned is not currently guaranteed, as there are known
        shortcomings that still need to be addressed, particularly with short stems and
        stems with flags attached.

        For rests, this simply returns ``ORIGIN``.
        """
        if self.rest:
            # Since there's no use-case we know of for rest tremolos, there's no
            # reasonable implementation of this, so just return 0, 0.
            return ORIGIN
        x = ZERO
        noteheads_rect = self.notehead_column_bounding_rect
        if not self.duration.display.requires_stem:
            x = noteheads_rect.x + (noteheads_rect.width / 2)
        if self.stem_direction == DirectionY.DOWN:
            y = noteheads_rect.y + noteheads_rect.height + self.staff.unit(1.5)
        else:
            y = noteheads_rect.y - self.staff.unit(1)
        return Point(x, y)

    @cached_property
    def notehead_column_bounding_rect(self) -> Rect:
        """The bounding rect of the notehead column after layout."""
        if self.rest:
            return Rect(ZERO, ZERO, ZERO, ZERO)
        left_x = Unit(float("inf"))
        top_y = Unit(float("inf"))
        right_x = Unit(float("-inf"))
        bottom_y = Unit(float("-inf"))
        for n in self.noteheads:
            n_rect = n.bounding_rect.offset(n.pos)
            if left_x > n_rect.x:
                left_x = n_rect.x
            if top_y > n_rect.y:
                top_y = n_rect.y
            n_rect_right_x = n_rect.x + n_rect.width
            if right_x < n_rect_right_x:
                right_x = n_rect_right_x
            n_rect_bottom_y = n_rect.y + n_rect.height
            if bottom_y < n_rect_bottom_y:
                bottom_y = n_rect_bottom_y
        return Rect(left_x, top_y, right_x - left_x, bottom_y - top_y)

    @cached_property
    def noteheads_outside_staff(self) -> List[Notehead]:
        """All noteheads which are above or below the staff"""
        return [
            note
            for note in self.noteheads
            # Since ``note.parent == self`` and ``self.y == 0``
            if not self.staff.y_inside_staff(note.y)
        ]

    @cached_property
    def leftmost_notehead_outside_staff(self) -> Optional[Notehead]:
        """The notehead furthest to the left outside the staff"""
        return min(self.noteheads_outside_staff, key=lambda n: n.x, default=None)

    @cached_property
    def rightmost_notehead_outside_staff(self) -> Optional[Notehead]:
        """The notehead furthest to the right outside the staff"""
        return max(self.noteheads_outside_staff, key=lambda n: n.x, default=None)

    @cached_property
    def notehead_column_outside_staff_width(self) -> Unit:
        """The total width of any noteheads outside the staff"""
        left_bounding_note = self.leftmost_notehead_outside_staff
        if not left_bounding_note:
            return ZERO
        else:
            extent = max((n.x + n.visual_width for n in self.noteheads_outside_staff))
            return extent - left_bounding_note.x

    # Can't use cached_property here because a setter is needed. Could get around this
    # by storing this computation in an attr like self._stem_direction though.
    @property
    def stem_direction(self) -> DirectionY:
        """The direction of the stem

        Takes the notehead furthest from the center of the staff, and returns the
        opposite direction.

        If the furthest notehead is in the center of the staff, the direction defaults
        to ``DirectionY.DOWN``, unless the staff has only one line, in which case it
        defaults to ``DirectionY.UP`` as a convenience for percussion staves.

        This automatically calculated property may be overridden using its setter. To
        revert to the automatically calculated value set this property to ``None``.

        If there are no noteheads (meaning this Chordrest is a rest), this arbitrarily
        returns ``DirectionY.UP``.

        """
        if self._stem_direction_override:
            return self._stem_direction_override
        furthest = self.furthest_notehead
        if furthest is None:
            return DirectionY.UP
        if furthest.y < self.staff.center_y:
            return DirectionY.DOWN
        elif furthest.y == self.staff.center_y:
            if self.staff.line_count == 1:
                return DirectionY.UP
            else:
                return DirectionY.DOWN
        else:
            return DirectionY.UP

    @stem_direction.setter
    def stem_direction(self, value: Optional[DirectionY]):
        self._stem_direction_override = value
        self._rebuild()

    @cached_property
    def stem_height(self) -> Unit:
        """The height of the stem"""
        flag_offset = self.staff.unit(Flag.vertical_offset_needed(self.duration))
        min_abs_height = self.staff.unit(2.5) + flag_offset
        fitted_abs_height = (
            abs(self.lowest_notehead.y - self.highest_notehead.y)
            + self.staff.unit(2.5)
            + flag_offset
        )
        abs_height = max(min_abs_height, fitted_abs_height)
        return abs_height

    def _clear(self):
        for notehead in self.noteheads:
            notehead.remove()
        self._noteheads = []
        for accidental in self.accidentals:
            accidental.remove()
        self._accidentals = []
        for ledger in self.ledgers:
            ledger.remove()
        self._ledgers = []
        for dot in self.dots:
            dot.remove()
        self._dots = []
        if self.stem:
            self.stem.remove()
        self._stem = None
        if self.flag:
            self.flag.remove()
        self._flag = None
        if self.rest:
            self.rest.remove()
        self._rest = None
        for prop in Chordrest._CACHED_PROPS:
            self.__dict__.pop(prop, None)

    def _rebuild(self):
        """Generate or regenerate all child objects"""
        if self.noteheads or self.rest:
            # Clear existing glyphs
            self._clear()
        if self.notes:
            for note in self.notes:
                if isinstance(note, tuple) and len(note) == 2:
                    pitch = note[0]
                    glyph_override = note[1]
                else:
                    pitch = note
                    glyph_override = None
                self._noteheads.append(
                    Notehead(
                        ZERO,
                        self,
                        pitch,
                        self.duration,
                        table=self.table,
                        glyph_override=glyph_override,
                    )
                )
            self._rest = None
            self._create_stem()
            if self.stem:
                self._position_noteheads_around_stem()
            self._create_accidentals()
            self._position_accidentals_horizontally()
            self._create_ledgers()
            self._create_flag()
        else:
            rest_y = self.rest_y or self.staff.center_y
            self._rest = Rest(Point(self.staff.unit(0), rest_y), self, self.duration)
        # Both rests and chords needs dots
        self._create_dots()

    def _create_ledgers(self):
        """Create all required ledger lines and store them in ``self.ledgers``

        This should be called after _position_noteheads_horizontally()
        as it relies on the position of noteheads in the chord to
        calculate the position and length of the ledger lines
        """
        pos_x = self.leftmost_notehead.x
        length = self.notehead_column_outside_staff_width
        for staff_pos in self.ledger_line_positions:
            self.ledgers.append(LedgerLine(Point(pos_x, staff_pos), self, length))

    def _create_accidentals(self):
        for notehead in self.noteheads:
            if notehead.pitch.accidental is None:
                continue
            accidental = Accidental(
                # X value is a placeholder to be resolved
                # in _position_accidentals_horizontally
                (ZERO, notehead.y),
                self,
                notehead.pitch.accidental,
            )
            self.accidentals.append(accidental)

    def _create_dots(self):
        """Create all the RhythmDots needed by this Chordrest."""
        for dot_pos in self.rhythm_dot_positions:
            self.dots.append(RhythmDot(dot_pos, self))

    def _create_stem(self):
        """If needed, create a Stem and store it in ``self.stem``."""
        if not self.duration.display.requires_stem:
            return
        if self.stem_direction == DirectionY.UP:
            attached_notehead = self.lowest_notehead
            anchor_key = "stemUpSE"
        else:
            attached_notehead = self.highest_notehead
            anchor_key = "stemDownNW"
        resolved_anchor_y = ZERO
        # Special case guard needed for invisible noteheads without music chars
        if attached_notehead.text:
            anchors = attached_notehead.music_chars[0].glyph_info.anchors
            if anchors:
                resolved_anchor = anchors.get(anchor_key)
                if resolved_anchor:
                    resolved_anchor_y = resolved_anchor.y
        self._stem = Stem(
            Point(ZERO, attached_notehead.y + resolved_anchor_y),
            self,
            self.stem_direction,
            self.stem_height,
        )

    def _create_flag(self):
        """Create a Flag attached to ``self.stem`` and store it in ``self.flag``"""
        if self.duration.display.flag_count:
            self._flag = Flag(
                (self.stem.pen.thickness / -2, ZERO),
                self.stem.end_point,
                self.duration,
                self.stem.direction,
            )

    def _position_noteheads_around_stem(self):
        """Reposition noteheads so that they are laid out correctly around the stem.

        This should only be run if a stem exists.
        """
        # Find the preferred side of the stem for noteheads,
        default_side = (
            DirectionX.LEFT
            if self.stem_direction == DirectionY.UP
            else DirectionX.RIGHT
        )
        # Start last staff pos at sentinel infinity position.
        # Rather than working with staff positions, we can work with
        # ``Notehead.y`` values directly because we know they all share
        # ``self`` as a parent.
        prev_y = Unit(float("inf"))
        # Start prev_side at wrong side so first note goes on the default side
        prev_side = default_side.flip()
        for note in sorted(
            self.noteheads,
            key=lambda n: n.y,
            reverse=(self.stem_direction == DirectionY.UP),
        ):
            if abs(prev_y - note.y) < self.staff.unit(1):
                # This note collides with previous, switch sides
                prev_side = prev_side.flip()
            else:
                prev_side = default_side
            note.x = self._resolve_notehead_x_pos(note, prev_side)
            prev_y = note.y

    def _resolve_notehead_x_pos(
        self, notehead: Notehead, stem_side: DirectionX
    ) -> Unit:
        if not notehead.text:
            return ZERO
        anchors = notehead.music_chars[0].glyph_info.anchors
        if not anchors:
            if stem_side == DirectionX.LEFT:
                return -notehead.visual_width
            return ZERO
        stem_offset = self.stem.pen.thickness / 2
        if stem_side == DirectionX.LEFT:
            anchor_key = "stemUpSE"
            stem_offset *= -1
        else:
            anchor_key = "stemDownNW"
        return -(anchors[anchor_key].x + stem_offset)

    def _position_accidentals_horizontally(self):
        """Reposition accidentals so that they are laid out correctly

        This implementation is very basic and fails in some fairly common use-cases. A
        proper solution would likely involve totally redoing this. See
        https://github.com/DigiScore/neoscore/issues/32
        """
        # Rather than working with staff positions, we can work with
        # ``Accidental.y`` values directly because we know they all share
        # ``self`` as a parent.
        padding = self.staff.music_font.engraving_defaults["staffLineThickness"]
        notehead_col_edge_x = self.leftmost_notehead.x
        prev_side = None
        prev_acc = None
        prev_rect = None
        for acc in sorted(self.accidentals, key=lambda a: a.y):
            acc_rect = acc.bounding_rect
            shift_for_collision = False

            if (
                prev_acc
                and prev_side == DirectionX.RIGHT
                and (prev_acc.y + prev_rect.y + prev_rect.height > acc.y + acc_rect.y)
            ):
                # This accidental collides with previous, switch sides
                shift_for_collision = True
                prev_side = DirectionX.LEFT
            else:
                prev_side = DirectionX.RIGHT
            if shift_for_collision:
                acc.x = prev_acc.x - acc.bounding_rect.width
            else:
                acc.x = notehead_col_edge_x - acc.bounding_rect.width - padding
            prev_acc = acc
            prev_rect = acc_rect



from __future__ import annotations

from typing import Optional

from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject


class RhythmDot(MusicText):

    """A single rhythmic augmentation dot used in notes and rests"""

    _glyph_name = "augmentationDot"

    def __init__(
        self, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ):
        """
        Args:
            pos: Position relative to ``parent``
            parent: The parent of the rest. If no font is provided,
                this parent or one of its ancestors must implement :obj:`.HasStaffUnit`.
            font: If provided, this overrides any font found in the ancestor chain.
        """
        MusicText.__init__(self, pos, parent, [self._glyph_name], font)



"""Objects representing individual elements of :obj:`.Path`\ s.

With the exception of ``CurveTo``, each of these classes is just a bare
``PositionedObject``.
"""


from __future__ import annotations

from neoscore.core.positioned_object import PositionedObject


class PathElement(PositionedObject):
    """Superclass for all path element types"""

    def __repr__(self) -> str:
        return f"{type(self).__name__}(pos={self.pos}, parent={self.parent})"


class MoveTo(PathElement):
    """An element representing the start of a new subpath"""


class LineTo(PathElement):
    """An element representing a straight line from the last path position"""


class ControlPoint(PathElement):
    """An element used as a control point in bezier curves.

    This element is not directly kept in :obj:`.Path.elements`, instead being placed
    inside :obj:`.CurveTo` elements.
    """


class CurveTo(PathElement):
    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        control_1: ControlPoint,
        control_2: ControlPoint,
    ):
        """
        Args:
            pos: The position of the object relative to its parent
            parent: The parent object.
            control_1: The curve's first control point
            control_2: The curve's second control point
        """
        super().__init__(pos, parent)
        self.control_1 = control_1
        self.control_2 = control_2

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(pos={self.pos}, parent={self.parent},"
            + f" c1={self.control_1}, c2={self.control_2})"
        )



from __future__ import annotations

from typing import Optional, Union

from neoscore.core.rect import Rect
from neoscore.core.units import Unit
from neoscore.interface.font_interface import FontInterface

_BOUNDING_RECT_CACHE: Dict[Tuple[Font, str], Rect] = {}


class Font:

    """A text font.

    All fonts are immutable. To get a modified version of a font, use
    :obj:`.Font.modified`
    """

    def __init__(
        self,
        family_name: str,
        size: Union[Unit, float],
        weight: Optional[int] = None,
        italic: bool = False,
    ):
        """
        Args:
            family_name: The font family name
            size: The size (height) of the font
            weight: The font weight on a 0-100 scale, where 50 is normal,
                lower numbers are lighter, and higher are darker.
                If ``None`` (the default), a normal weight will be used.
            italic: Whether the font is italicized
        """
        self._family_name = family_name
        self._size = size if isinstance(size, Unit) else Unit(size)
        self._weight = weight
        self._italic = italic
        self._interface = FontInterface(
            self.family_name, self.size, self.weight, self.italic
        )

    @property
    def family_name(self) -> str:
        return self._family_name

    @property
    def size(self) -> Unit:
        return self._size

    @property
    def weight(self) -> Optional[int]:
        return self._weight

    @property
    def italic(self) -> bool:
        return self._italic

    @property
    def ascent(self) -> Unit:
        """The ascent of the font.

        The ascent is the vertical distance between the font baseline and
        the highest any font characters reach.
        """
        return self._interface.ascent

    @property
    def descent(self) -> Unit:
        """The descent of the font.

        The ascent is the vertical distance between the font baseline and
        the lowest any font characters reach.
        """
        return self._interface.descent

    @property
    def x_height(self) -> Unit:
        """The x-height for the font."""
        return self._interface.x_height

    @property
    def interface(self) -> FontInterface:
        """The backing low-level font interface"""
        return self._interface

    def __str__(self):
        return f"Font('{self.family_name}', {self.size}, {self.weight}, {self.italic})"

    def __eq__(self, other):
        return (
            isinstance(other, Font)
            and self.family_name == other.family_name
            and self.size == other.size
            and self.weight == other.weight
            and self.italic == other.italic
        )

    def __hash__(self):
        return hash(
            (self.family_name, self.size.rounded_base_value, self.weight, self.italic)
        )

    def modified(
        self,
        family_name: Optional[str] = None,
        size: Optional[Union[Unit, float]] = None,
        weight: Optional[int] = None,
        italic: Optional[bool] = None,
    ) -> Font:
        """Derive a font from this one.

        All properties not specified will be taken from the existing font.
        """
        return Font(
            family_name if family_name is not None else self.family_name,
            size if size is not None else self.size,
            weight if weight is not None else self.weight,
            italic if italic is not None else self.italic,
        )

    def bounding_rect_of(self, string: str) -> Rect:
        """Approximate the bounding rect of a string in this font."""
        key = (self, string)
        cached_rect = _BOUNDING_RECT_CACHE.get(key)
        if cached_rect:
            return cached_rect
        rect = self._interface.bounding_rect_of(string)
        _BOUNDING_RECT_CACHE[key] = rect
        return rect



from PyQt5 import QtGui, QtWidgets
from PyQt5.QtCore import QPointF

from neoscore.core.key_event import KeyEventType
from neoscore.core.mouse_event import MouseEventType
from neoscore.interface.qt.converters import (
    q_key_event_to_key_event,
    q_mouse_event_to_mouse_event,
)

_NO_DRAG = 0
_SCROLL_HAND_DRAG = 1
_NO_VIEWPORT_UPDATE = 3
_SCROLL_BAR_AS_NEEDED = 0
_SCROLL_BAR_ALWAYS_OFF = 1
_FOCUS_POLICY_STRONG_FOCUS = 0x1 | 0x2 | 0x8
_FOCUS_POLICY_NO_FOCUS = 0


class Viewport(QtWidgets.QGraphicsView):
    """A QGraphicsView configured for use in interactive neoscore scenes.

    Includes some basic hacky features.
    """

    def __init__(self, scene: QtWidgets.QGraphicsScene):
        super().__init__(scene)
        # Default configs
        self.setViewport(QtWidgets.QOpenGLWidget())
        self.setRenderHint(QtGui.QPainter.Antialiasing)
        # Automatic viewport updates are disabled. Updates are performed
        # manually in the main window refresh function.
        self.set_auto_interaction(True)
        self.setViewportUpdateMode(_NO_VIEWPORT_UPDATE)  # noqa
        self.mouse_event_handler = None
        self.key_event_handler = None

    def set_auto_interaction(self, enabled: bool):
        """Set whether mouse and scrollbar interaction is enabled."""
        self.auto_interaction_enabled = enabled
        if enabled:
            self.setDragMode(_SCROLL_HAND_DRAG)  # noqa
            self.setHorizontalScrollBarPolicy(_SCROLL_BAR_AS_NEEDED)  # noqa
            self.setVerticalScrollBarPolicy(_SCROLL_BAR_AS_NEEDED)  # noqa
        else:
            self.setDragMode(_NO_DRAG)  # noqa
            self.setHorizontalScrollBarPolicy(_SCROLL_BAR_ALWAYS_OFF)  # noqa
            self.setVerticalScrollBarPolicy(_SCROLL_BAR_ALWAYS_OFF)  # noqa

    def wheelEvent(self, event):
        """Implementation of Qt event hook for zooming with the mouse wheel."""
        if not self.auto_interaction_enabled:
            return
        zoom_in_factor = 1.25
        zoom_out_factor = 1 / zoom_in_factor
        wheel_delta = event.angleDelta().y()
        # Save the scene pos
        old_pos = self.mapToScene(event.pos())
        # Zoom
        if wheel_delta > 0:
            zoom_factor = zoom_in_factor
        else:
            zoom_factor = zoom_out_factor
        self.scale(zoom_factor, zoom_factor)
        # Get the new position
        new_pos = self.mapToScene(event.pos())
        # Move scene to old position
        delta = new_pos - old_pos
        self.translate(delta.x(), delta.y())

    def scrollContentsBy(self, *args):
        """Override of superclass scroll action to trigger a viewport update."""
        super().scrollContentsBy(*args)
        self.viewport().update()

    def window_document_pos(self) -> QPointF:
        return self.mapToScene(0, 0)

    # Input event handler overrides

    def mouseMoveEvent(self, e):
        if self.mouse_event_handler:
            self.mouse_event_handler(
                q_mouse_event_to_mouse_event(
                    e, MouseEventType.MOVE, self.window_document_pos()
                )
            )
        if self.auto_interaction_enabled:
            super().mouseMoveEvent(e)

    def mousePressEvent(self, e):
        if self.mouse_event_handler:
            self.mouse_event_handler(
                q_mouse_event_to_mouse_event(
                    e, MouseEventType.PRESS, self.window_document_pos()
                )
            )
        if self.auto_interaction_enabled:
            super().mousePressEvent(e)

    def mouseReleaseEvent(self, e):
        if self.mouse_event_handler:
            self.mouse_event_handler(
                q_mouse_event_to_mouse_event(
                    e, MouseEventType.RELEASE, self.window_document_pos()
                )
            )
        if self.auto_interaction_enabled:
            super().mouseReleaseEvent(e)

    def mouseDoubleClickEvent(self, e):
        if self.mouse_event_handler:
            self.mouse_event_handler(
                q_mouse_event_to_mouse_event(
                    e, MouseEventType.DOUBLE_CLICK, self.window_document_pos()
                )
            )
        if self.auto_interaction_enabled:
            super().mouseDoubleClickEvent(e)

    def keyPressEvent(self, e):
        if self.key_event_handler:
            self.key_event_handler(q_key_event_to_key_event(e, KeyEventType.PRESS))
        if self.auto_interaction_enabled:
            super().keyPressEvent(e)

    def keyReleaseEvent(self, e):
        if self.key_event_handler:
            self.key_event_handler(q_key_event_to_key_event(e, KeyEventType.RELEASE))
        if self.auto_interaction_enabled:
            super().keyPressEvent(e)

    # End of input event handler overrides



from __future__ import annotations

from typing import Optional, cast

from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.point import ORIGIN, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.spanner_2d import Spanner2D


class PedAndStar(Spanner2D, PositionedObject, HasMusicFont):

    """Pedal notation in the ornate 'Ped' and release star style."""

    def __init__(
        self,
        start: PointDef,
        start_parent: PositionedObject,
        end_pos: PointDef,
        end_parent: Optional[PositionedObject] = None,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            start: The position of the start-pedal mark relative to ``start_parent``.
            start_parent: Anchor for the start-pedal mark.
            end_pos: The position of the release-pedal mark relative to ``end_parent``.
            end_parent: An optional anchor for the release-pedal mark.
                Defaults to ``self``.
            font: If provided, this overrides any font found in the ancestor chain.
        """
        PositionedObject.__init__(self, start, start_parent)
        Spanner2D.__init__(
            self,
            end_pos,
            cast(PositionedObject, end_parent) if end_parent else self,
        )

        if font is None:
            font = HasMusicFont.find_music_font(start_parent)
        self._music_font = font

        # Add opening pedal mark
        self.depress_mark = MusicText(ORIGIN, self, "keyboardPedalPed", font)
        self.lift_mark = MusicText(
            self.end_pos, self.end_parent, "keyboardPedalUp", font
        )

    @property
    def music_font(self) -> MusicFont:
        return self._music_font



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, cast

from neoscore.core.exceptions import NoAncestorStaffError
from neoscore.core.point import Point
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import Unit

if TYPE_CHECKING:
    from neoscore.western.abstract_staff import AbstractStaff


class StaffObject:

    """An object which must always be the descendant of an :obj:`.AbstractStaff`.

    This is a mixin class for :obj:`.PositionedObject` classes.
    """

    def __init__(self, parent: PositionedObject):
        staff = StaffObject._find_staff(cast(PositionedObject, parent))
        if not staff:
            raise NoAncestorStaffError
        self._staff = staff

    @property
    def staff(self) -> AbstractStaff:
        """The ancestor staff"""
        return self._staff

    @property
    def pos_in_staff(self) -> Point:
        """The logical position of this object relative to the staff."""
        return self.staff.map_to(cast(PositionedObject, self))

    @property
    def pos_x_in_staff(self) -> Unit:
        """A specialized version of ``pos_in_staff`` which only finds the x pos"""
        return self.staff.map_x_to(cast(PositionedObject, self))

    @staticmethod
    def _find_staff(obj: PositionedObject) -> Optional[AbstractStaff]:
        """Find the first staff which is an ancestor of ``obj`` or ``obj`` itself"""
        marker = "_neoscore_abstract_staff_type_marker"
        if hasattr(obj, marker):
            return cast(Any, obj)
        return obj.first_ancestor_with_attr(marker)



"""Helper methods for conversion between ``neoscore.core`` and Qt classes"""

from typing import Union

from PyQt5.QtCore import QPoint, QPointF, QRect, QRectF
from PyQt5.QtGui import QColor, QKeyEvent, QMouseEvent

from neoscore.core.color import Color
from neoscore.core.key_event import KeyEvent, KeyEventType
from neoscore.core.mouse_event import MouseButton, MouseEvent, MouseEventType
from neoscore.core.point import Point
from neoscore.core.rect import Rect
from neoscore.core.units import Unit


def qt_point_to_point(qt_point: Union[QPoint, QPointF]) -> Point:
    """Create a Point from a QPoint or QPointF

    Args:
        qt_point: The source point
    """
    return Point(Unit(qt_point.x()), Unit(qt_point.y()))


def point_to_qt_point(point: Point) -> QPoint:
    """Create a QPoint from a Point

    Args:
        point: The source point
    """
    return QPoint(int(point.x.base_value), int(point.y.base_value))


def point_to_qt_point_f(point: Point) -> QPointF:
    """Create a QPointF from a Point

    Args:
        point: The source point
    """
    return QPointF(point.x.base_value, point.y.base_value)


def qt_rect_to_rect(qt_rect: Union[QRect, QRectF]) -> Rect:
    """Create a Rect from a QRect or QRectF

    Args:
        qt_rect: The source rect
    """
    return Rect(
        Unit(qt_rect.x()),
        Unit(qt_rect.y()),
        Unit(qt_rect.width()),
        Unit(qt_rect.height()),
    )


def rect_to_qt_rect(rect: Rect) -> QRect:
    """Create a QRect from a Rect

    Args:
        rect: The source rect
    """
    return QRect(
        int(rect.x.base_value),
        int(rect.y.base_value),
        int(rect.width.base_value),
        int(rect.height.base_value),
    )


def rect_to_qt_rect_f(rect: Rect) -> QRectF:
    """Create a QRectF from a Rect

    Args:
        rect: The source rect
    """
    return QRectF(
        rect.x.base_value,
        rect.y.base_value,
        rect.width.base_value,
        rect.height.base_value,
    )


def color_to_q_color(color: Color) -> QColor:
    """Create a ``QColor`` from a ``Color``

    Args:
        color: The source ``Color``
    """
    return QColor(color.red, color.green, color.blue, color.alpha)


def q_color_to_color(q_color: QColor) -> Color:
    """Create a ``Color`` from a ``QColor``

    Args:
        q_color: The source ``QColor``
    """
    return Color(q_color.red(), q_color.green(), q_color.blue(), q_color.alpha())


_Q_MOUSE_LEFT_BUTTON = 0x00000001
_Q_MOUSE_RIGHT_BUTTON = 0x00000002
_Q_MOUSE_MIDDLE_BUTTON = 0x00000004


def q_mouse_event_to_mouse_event(
    q_event: QMouseEvent, ns_event_type: MouseEventType, window_pos: QPointF
) -> MouseEvent:
    buttons = int(q_event.buttons())
    if buttons & _Q_MOUSE_LEFT_BUTTON:
        ns_mouse_button = MouseButton.LEFT
    elif buttons & _Q_MOUSE_RIGHT_BUTTON:
        ns_mouse_button = MouseButton.RIGHT
    elif buttons & _Q_MOUSE_MIDDLE_BUTTON:
        ns_mouse_button = MouseButton.MIDDLE
    else:
        ns_mouse_button = None
    q_pos = q_event.windowPos()
    ns_window_pos = (int(q_pos.x()), int(q_pos.y()))
    ns_document_pos = Point(
        Unit(q_pos.x() + window_pos.x()), Unit(q_pos.y() + window_pos.y())
    )
    return MouseEvent(ns_event_type, ns_mouse_button, ns_window_pos, ns_document_pos)


def q_key_event_to_key_event(
    q_event: QKeyEvent, basic_event_type: KeyEventType
) -> KeyEvent:
    resolved_event_type = (
        KeyEventType.AUTO_REPEAT if q_event.isAutoRepeat() else basic_event_type
    )
    return KeyEvent(
        resolved_event_type, q_event.key(), int(q_event.modifiers()), q_event.text()
    )



from __future__ import annotations

from math import atan, cos, pi, sin, sqrt, tan
from typing import List, Optional, Tuple, cast

from neoscore.core.brush import Brush, BrushDef
from neoscore.core.layout_controllers import NewLine
from neoscore.core.painted_object import PaintedObject
from neoscore.core.path_element import (
    ControlPoint,
    CurveTo,
    LineTo,
    MoveTo,
    PathElement,
)
from neoscore.core.pen import Pen, PenDef
from neoscore.core.point import ORIGIN, Point, PointDef
from neoscore.core.positioned_object import PositionedObject, render_cached_property
from neoscore.core.units import ZERO, Mm, Unit
from neoscore.interface.path_interface import (
    PathInterface,
    ResolvedCurveTo,
    ResolvedLineTo,
    ResolvedMoveTo,
    ResolvedPathElement,
)

_TWO_PI = pi * 2
_HALF_PI = pi / 2


class Path(PaintedObject):

    """A vector path whose elements can be anchored to other objects.

    If a ``Path`` is in a :obj:`.Flowable`, any element parents in the path should be in the
    same flowable. Likewise, if a path is not in a flowable, all element parts should
    not be in one either.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        rotation: float = 0,
        background_brush: Optional[BrushDef] = None,
        transform_origin: PointDef = ORIGIN,
    ):
        """
        Args:
            pos: The position of the path root.
            parent: The parent object or None
            brush: The brush to fill shapes with.
            pen: The pen to draw outlines with.
            rotation: Angle in degrees. Rotated paths with flowable breaks or
                path elements parented to other objects are not currently supported.
            background_brush: Optional brush used to paint the path's bounding rect
                behind it.
            transform_origin: The origin point for rotation and scaling transforms
        """
        super().__init__(pos, parent, brush, pen)
        self.background_brush = background_brush
        self._rotation = rotation
        self.elements: List[PathElement] = []
        self._current_subpath_start: Optional[Tuple[Point, Optional[parent]]] = None
        self.transform_origin = transform_origin

    @classmethod
    def straight_line(
        cls,
        start: PointDef,
        parent: Optional[PositionedObject],
        end: PointDef,
        end_parent: Optional[PositionedObject] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ) -> Path:
        """Convenience for drawing a single straight line.

        Args:
            start: The position of the center of the line's start
            parent: A parent object
            end: The position of the end of the line, relative to ``end_parent``
                if provided or otherwise ``start``
            end_parent: An optional parent for the end point.
            brush: The brush to fill shapes with.
            pen: The pen to draw outlines with. Defaults to no pen.
        """
        line = cls(start, parent, brush, pen)
        end = Point.from_def(end)
        line.line_to(end.x, end.y, end_parent)
        return line

    @classmethod
    def rect(
        cls,
        pos: PointDef,
        parent: Optional[PositionedObject],
        width: Unit,
        height: Unit,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """Convenience for drawing a rectangle."""
        path = cls(pos, parent, brush, pen)
        path.line_to(width, ZERO)
        path.line_to(width, height)
        path.line_to(ZERO, height)
        path.line_to(ZERO, ZERO)
        return path

    @classmethod
    def ellipse(
        cls,
        pos: PointDef,
        parent: Optional[PositionedObject],
        width: Unit,
        height: Unit,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """Convenience for drawing an ellipse.

        ``pos`` indicates the top left corner of the ellipse's bounding rect.
        To create a path from a center point, use ``Path.ellipse_from_center``.

        This can also be used for drawing circles by giving the same width and height.
        """
        # Algorithm courtesy of https://stackoverflow.com/a/2173084/5615927
        path = cls(pos, parent, brush, pen)
        kappa = 0.5522848
        ox = (width / 2) * kappa  # control point offset horizontal
        oy = (height / 2) * kappa  # control point offset vertical
        xe = width  # x-end
        ye = height  # y-end
        xm = width / 2  # x-middle
        ym = height / 2  # y-middle
        path.move_to(ZERO, ym)
        path.cubic_to(ZERO, ym - oy, xm - ox, ZERO, xm, ZERO)
        path.cubic_to(xm + ox, ZERO, xe, ym - oy, xe, ym)
        path.cubic_to(xe, ym + oy, xm + ox, ye, xm, ye)
        path.cubic_to(xm - ox, ye, ZERO, ym + oy, ZERO, ym)
        return path

    @classmethod
    def ellipse_from_center(
        cls,
        center_pos: PointDef,
        parent: Optional[PositionedObject],
        width: Unit,
        height: Unit,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """Convenience for drawing an ellipse from its center point.

        The constructed path will have its ``pos`` at the ellipse bounding rect's top
        left corner.

        See also ``Path.ellipse``
        """
        center_pos = Point.from_def(center_pos)
        return Path.ellipse(
            Point(center_pos.x - (width / 2), center_pos.y - (height / 2)),
            parent,
            width,
            height,
            brush,
            pen,
        )

    @classmethod
    def arc(
        cls,
        pos: PointDef,
        parent: Optional[PositionedObject],
        width: Unit,
        height: Unit,
        start_angle: float,
        stop_angle: float,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """Convenience for drawing an elliptical arc.

        Args:
            pos: The position of the upper left corner of the traced ellipse. parent:
                The parent object or None width: The traced ellipse's width height: The
                traced ellipse's height start_angle: The starting arc angle in radians
                clockwise relative
                to the 3 o'clock position.
            parent: The parent object or None
            width: The full ellipse width
            height: The full ellipse height
            start_angle: The start arc angle in radians clockwise relative
                to the 3 o'clock position.
            stop_angle: The stopping arc angle, defined like ``start_angle``.
            brush: The brush to fill shapes with. pen: The pen to draw outlines with.
            pen: The pen to draw outlines with.

        The arc definition can be most easily understood as tracing an ellipse as
        defined in ``Path.ellipse()``, where ``pos`` marks the top-left corner of the
        ellipse bounding rect. Two angles are provided in clockwise radians relative to
        the 3 o'clock position. The arc is traced from ``start_angle`` clockwise to
        ``stop_angle``. Consequently, depending on the provided angles the actually
        drawn path may be far from the Path's position.

        The provided angles are interpreted mod ``2*pi``. The angle between them must
        not be 0. This also means arc angles of ``2*pi``, (i.e. complete ellipses), are
        not supported as they are interpreted as 0. Complete ellipses should instead be
        drawn with ``Path.ellipse()``.

        Raises:
            ValueError: if invalid angles are given
        """
        # This method and ``_acute_arc_to_bezier`` are adapted from Joe
        # Cridge's algorithm and JS implementation found at
        # https://www.joecridge.me/bezier.pdf. Most of the original
        # comments have been left in place as well.

        # Work in float base units, converting back to units at the end
        w = width.base_value
        h = height.base_value

        # Make all angles positive...
        while start_angle < 0:
            start_angle += _TWO_PI
        while stop_angle < 0:
            stop_angle += _TWO_PI

        # ...and confine them to the interval [0,TWO_PI).
        start_angle %= _TWO_PI
        stop_angle %= _TWO_PI

        # Adjust angles to counter linear scaling.
        if start_angle <= _HALF_PI:
            start_angle = atan(w / h * tan(start_angle))
        elif _HALF_PI < start_angle <= 3 * _HALF_PI:
            start_angle = atan(w / h * tan(start_angle)) + pi
        else:
            start_angle = atan(w / h * tan(start_angle)) + _TWO_PI
        if stop_angle <= _HALF_PI:
            stop_angle = atan(w / h * tan(stop_angle))
        elif _HALF_PI < stop_angle <= 3 * _HALF_PI:
            stop_angle = atan(w / h * tan(stop_angle)) + pi
        else:
            stop_angle = atan(w / h * tan(stop_angle)) + _TWO_PI

        # Exceed the interval if necessary in order to preserve the size and
        # orientation of the arc.
        if start_angle > stop_angle:
            stop_angle += _TWO_PI

        # Create curves
        epsilon = 0.00001  # Smallest visible angle on displays up to 4K.
        curves = []
        while stop_angle - start_angle > epsilon:
            arc_to_draw = min(stop_angle - start_angle, _HALF_PI)
            curves.append(Path._acute_arc_to_bezier(start_angle, arc_to_draw))
            start_angle += arc_to_draw

        if not len(curves):
            raise ValueError(f"Invalid arc angles {start_angle} and {stop_angle}.")

        # Draw curves
        path = cls(pos, parent, brush, pen)
        rx = w / 2.0
        ry = h / 2.0
        path.move_to(Unit(rx * curves[0]["ax"] + rx), Unit(ry * curves[0]["ay"] + ry))
        for curve in curves:
            path.cubic_to(
                Unit(rx * curve["bx"] + rx),
                Unit(ry * curve["by"] + ry),
                Unit(rx * curve["cx"] + rx),
                Unit(ry * curve["cy"] + ry),
                Unit(rx * curve["dx"] + rx),
                Unit(ry * curve["dy"] + ry),
            )
        return path

    @staticmethod
    def _acute_arc_to_bezier(start: float, size: float) -> dict:
        """
        Generate a cubic Bezier representing an arc on the unit circle of total
        angle ``size`` radians, beginning ``start`` radians above the x-axis.
        """
        # Evaluate constants.
        alpha = size / 2.0
        cos_alpha = cos(alpha)
        sin_alpha = sin(alpha)
        cot_alpha = 1.0 / tan(alpha)
        phi = start + alpha  # This is how far the arc needs to be rotated.
        cos_phi = cos(phi)
        sin_phi = sin(phi)
        lambda_ = (4.0 - cos_alpha) / 3.0
        mu = sin_alpha + (cos_alpha - lambda_) * cot_alpha
        # Return rotated waypoints.
        return {
            "ax": cos(start),
            "ay": sin(start),
            "bx": lambda_ * cos_phi + mu * sin_phi,
            "by": lambda_ * sin_phi - mu * cos_phi,
            "cx": lambda_ * cos_phi - mu * sin_phi,
            "cy": lambda_ * sin_phi + mu * cos_phi,
            "dx": cos(start + size),
            "dy": sin(start + size),
        }

    @classmethod
    def arrow(
        cls,
        start: PointDef,
        parent: Optional[PositionedObject],
        end: PointDef,
        end_parent: Optional[PositionedObject] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        line_width: Unit = Mm(0.35),
        arrow_head_width: Unit = Mm(1.5),
        arrow_head_length: Unit = Mm(2.5),
    ) -> Path:
        """Convenience for drawing an arrow

        Args:
            start: The position of the center of the arrow line's start
            parent: A parent object
            end: The position of the arrow's tip, relative to ``end_parent``
                if provided or ``start``
            end_parent: An optional parent for the end point.
            brush: The brush to fill shapes with.
            pen: The pen to draw outlines with. Defaults to no pen.
            line_width: The thickness of the arrow's line
            arrow_head_width: The width of the arrow head extending perpendicular
                from the line.
            arrow_head_length: The length of the arrow head parallel to the line.

        Note that ``end_parent`` is only used for initially drawing the
        path. If ``end_parent`` moves relative to the path after
        creation, the path shape will not be automatically updated.
        """
        # This algorithm is lightly adapted from
        # https://github.com/frogcat/canvas-arrow by Yuzo Matsuzawa
        # released under MIT.
        path = cls(start, parent, brush, pen or Pen.no_pen())
        end = Point.from_def(end)
        if end_parent:
            end = path.map_to(end_parent) + end
        # The original algorithm supports diverse shapes, including
        # things like double-headed arrows, using this system of a
        # variable numbers of control points. We don't use them right
        # now, but since we might want to later (and since I don't
        # understand the code enough to refactor it) it's left in
        # place here.
        control_points = [
            (0, line_width.base_value / 2),
            (-arrow_head_length.base_value, line_width.base_value / 2),
            (-arrow_head_length.base_value, arrow_head_width.base_value / 2),
        ]
        # The start pos (relative to the path) is always (0, 0), so dx
        # and dy are just the end position
        dx = end.x.base_value
        dy = end.y.base_value
        length = sqrt(dx * dx + dy * dy)
        sin_ = dy / length
        cos_ = dx / length
        resolved_points: List[Tuple[float, float]] = [(0, 0)]
        for cp in control_points:
            if cp[0] < 0:
                resolved_points.append((length + cp[0], cp[1]))
            else:
                resolved_points.append(cp)
        resolved_points.append((length, 0))
        for cp in reversed(control_points):
            if cp[0] < 0:
                resolved_points.append((length + cp[0], -cp[1]))
            else:
                resolved_points.append((cp[0], -cp[1]))
        resolved_points.append((0, 0))
        path.move_to(Unit(resolved_points[0][0]), Unit(resolved_points[0][1]))
        for rp in resolved_points[1:]:
            x = rp[0] * cos_ - rp[1] * sin_
            y = rp[0] * sin_ + rp[1] * cos_
            path.line_to(Unit(x), Unit(y))
        return path

    @render_cached_property
    def _cacheable_breakable_length(self) -> Unit:
        # Find the positions of every path element relative to the path
        min_x = Unit(float("inf"))
        max_x = Unit(-float("inf"))
        for element in self.elements:
            # Determine element X relative to self
            relative_x = self.map_x_to(element)
            # Now update min/max accordingly
            if relative_x > max_x:
                max_x = relative_x
            if relative_x < min_x:
                min_x = relative_x
        return max_x - min_x

    @property
    def breakable_length(self) -> Unit:
        """The breakable length of the path.

        This is calculated automatically from path contents. By extension,
        this means that by default all ``Path`` objects will automatically
        wrap in flowables.
        """
        # This can be slow to calculate, so we want to render-cache it.
        # We can't directly annotate this method @render_cached_property
        # because it would violate the type signature of this inherited method,
        # so simply delegate to a private property which is cached.
        return self._cacheable_breakable_length

    @property
    def background_brush(self) -> Optional[Brush]:
        """An optional brush to paint over the path bounding rect's background with"""
        return self._background_brush

    @background_brush.setter
    def background_brush(self, value: Optional[BrushDef]):
        if value:
            self._background_brush = Brush.from_def(value)
        else:
            self._background_brush = None

    def line_to(self, x: Unit, y: Unit, parent: Optional[PositionedObject] = None):
        """Draw a path from the current position to a new point.

        If the path is empty, this will add two elements, an initial
        ``MoveTo(ORIGIN, self)`` and the requested ``LineTo``.

        Args:
            x: The end x position
            y: The end y position
            parent: An optional parent, whose position the target coordinate
                will be relative to.

        """
        if not len(self.elements):
            self.move_to(ZERO, ZERO)
        self.elements.append(LineTo(Point(x, y), parent or self))

    def move_to(self, x: Unit, y: Unit, parent: Optional[PositionedObject] = None):
        """Close the current sub-path and start a new one.

        Args:
            x: The end x position
            y: The end y position
            parent: An optional parent, whose position the target coordinate will
                be relative to.
        """
        self._current_subpath_start = (Point(x, y), parent or self)
        self.elements.append(MoveTo(Point(x, y), parent or self))

    def close_subpath(self):
        """Close the current sub-path with a line.

        Draw a line back to the starting position (including any parent) of the current
        subpath.
        """
        end_pos = Point.from_def(self._current_subpath_start[0])
        end_parent = self._current_subpath_start[1]
        self.line_to(
            end_pos.x,
            end_pos.y,
            end_parent,
        )

    def cubic_to(
        self,
        control_1_x: Unit,
        control_1_y: Unit,
        control_2_x: Unit,
        control_2_y: Unit,
        end_x: Unit,
        end_y: Unit,
        control_1_parent: Optional[PositionedObject] = None,
        control_2_parent: Optional[PositionedObject] = None,
        end_parent: Optional[PositionedObject] = None,
    ):
        """Draw a cubic bezier curve from the current position to a new point.

        If the path is empty, this will add two elements, an initial
        ``MoveTo(ORIGIN, self)`` and the requested ``CurveTo``.

        Args:
            control_1_x: The x coordinate of the first control point.
            control_1_y: The y coordinate of the first control point.
            control_2_x: The x coordinate of the second control point.
            control_2_y: The y coordinate of the second control point.
            end_x: The x coordinate of the curve target.
            end_y: The y coordinate of the curve target.
            control_1_parent: An optional parent for
                the first control point. Defaults to ``self``.
            control_2_parent: An optional parent for
                the second control point. Defaults to ``self``.
            end_parent: An optional parent for the
                curve target. Defaults to ``self``.

        """
        c1 = ControlPoint(
            Point(control_1_x, control_1_y),
            control_1_parent or self,
        )
        c2 = ControlPoint(
            Point(control_2_x, control_2_y),
            control_2_parent or self,
        )
        if not len(self.elements):
            self.move_to(ZERO, ZERO)
        self.elements.append(CurveTo(Point(end_x, end_y), end_parent or self, c1, c2))

    def _relative_element_pos(self, element: PositionedObject) -> Point:
        return self.map_to(element)

    def _resolve_path_elements(self) -> List[ResolvedPathElement]:
        resolved: List[ResolvedPathElement] = []
        for element in self.elements:
            # Interface drawing methods expect coordinates
            # relative to PathInterface root
            pos = self._relative_element_pos(element)
            if isinstance(element, LineTo):
                resolved.append(ResolvedLineTo(pos.x, pos.y))
            elif isinstance(element, MoveTo):
                resolved.append(ResolvedMoveTo(pos.x, pos.y))
            elif isinstance(element, CurveTo):
                element = cast(CurveTo, element)
                resolved_c1_pos = self._relative_element_pos(element.control_1)
                resolved_c2_pos = self._relative_element_pos(element.control_2)
                resolved.append(
                    ResolvedCurveTo(
                        resolved_c1_pos.x,
                        resolved_c1_pos.y,
                        resolved_c2_pos.x,
                        resolved_c2_pos.y,
                        pos.x,
                        pos.y,
                    )
                )
            else:
                raise TypeError("Unknown PathElement type")
        return resolved

    def _render_slice(
        self,
        pos: Point,
        inside_flowable: bool,
        clip_start_x: Optional[Unit] = None,
        clip_width: Optional[Unit] = None,
    ):
        # If this proves to be a performance bottleneck in the future,
        # it is very possible to optimize this to create `PathInterface`s
        # which reuse `QPainterPath`s.
        resolved_path_elements = self._resolve_path_elements()
        slice_interface = PathInterface(
            pos,
            None if inside_flowable else self.parent.interface_for_children,
            self.scale,
            self.rotation,
            self.transform_origin,
            self.brush.interface,
            self.pen.interface,
            resolved_path_elements,
            self.background_brush.interface if self.background_brush else None,
            clip_start_x,
            clip_width,
        )
        slice_interface.render()
        self.interfaces.append(slice_interface)

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        inside_flowable = bool(flowable_line)
        self._render_slice(pos, inside_flowable, None, None)

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        slice_length = flowable_line.length - (flowable_x - flowable_line.flowable_x)
        self._render_slice(pos, True, ZERO, slice_length)

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        self._render_slice(pos, True, object_x, flowable_line.length)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        self._render_slice(pos, True, object_x, None)



from __future__ import annotations

from typing import Optional

from neoscore.core.brush import BrushDef
from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.music_font import MusicFont
from neoscore.core.path import Path
from neoscore.core.pen import PenDef
from neoscore.core.point import ORIGIN, PointDef
from neoscore.core.positioned_object import PositionedObject


class MusicPath(Path, HasMusicFont):
    """A Path with a music font.

    This is mostly useful as a superclass for paths which use a
    MusicFont's engraving defaults and staff unit to control their
    appearance, but it can also be instantiated directly.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject] = None,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        rotation: float = 0,
        background_brush: Optional[BrushDef] = None,
        transform_origin: PointDef = ORIGIN,
    ):
        """
        Args:
            pos: The position of the path root.
            parent: The parent object. If no font is given, this or one of its
                ancestors must implement ``HasMusicFont``.
            font: If provided, this overrides any font found in the ancestor chain.
            brush: The brush to fill shapes with.
            pen: The pen to draw outlines with.
            rotation: Angle in degrees. Rotated paths with flowable breaks and
                path elements anchored to other objects are not currently supported.
            background_brush: Optional brush used to paint the path's bounding rect
                behind it.
            transform_origin: The origin point for rotation and scaling transforms
        """
        Path.__init__(
            self,
            pos,
            parent,
            brush,
            pen,
            rotation,
            background_brush,
            transform_origin,
        )
        if font is None:
            font = HasMusicFont.find_music_font(parent)
        self._music_font = font

    @property
    def music_font(self) -> MusicFont:
        return self._music_font



from __future__ import annotations

from typing import Optional

from neoscore.core import neoscore
from neoscore.core.brush import BrushDef
from neoscore.core.layout_controllers import NewLine
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.pen import PenDef
from neoscore.core.point import Point
from neoscore.core.text_alignment import AlignmentY
from neoscore.core.units import ZERO, Unit
from neoscore.western.staff_object import StaffObject
from neoscore.western.tab_staff import TabStaff


class TabClef(MusicText, StaffObject):

    """A tablature clef.

    Unlike classical :obj:`.Clef`\ s, this is purely cosmetic. It must be placed in a
    :obj:`.TabStaff`, and it is automatically positioned at its beginning. If the
    ``TabStaff`` is in a flowable, this automatically repeats at the beginning of every
    flowed staff line for the length of the staff. Because clef changes are generally
    inapplicable to tabs, clef changes are not currently supported.
    """

    # Type sentinel used to hackily check type
    # without importing the type, risking cyclic imports.
    _neoscore_tab_clef_type_marker = True

    def __init__(
        self,
        staff: TabStaff,
        glyph_name: str = "6stringTabClef",
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        hide_background: bool = True,
    ):
        """
        Args:
            staff: The parent staff
            glyph_name: The SMuFL glyph to use.
            font: The font to use. Defaults to the staff's font.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
            hide_background: Whether to paint over the background behind the text.
        """
        StaffObject.__init__(self, staff)
        MusicText.__init__(
            self,
            (ZERO, staff.center_y),
            staff,
            glyph_name,
            font,
            brush,
            pen,
            background_brush=neoscore.background_brush if hide_background else None,
            alignment_y=AlignmentY.CENTER,
        )

    @property
    def breakable_length(self) -> Unit:
        """Tab clefs are drawn at the beginning of every line in a staff."""
        return self.parent.breakable_length - self.x

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        if fringe_layout.pos_x_in_staff == self.pos_x_in_staff:
            pos = Point(pos.x + fringe_layout.clef, pos.y)
        super().render_complete(pos, flowable_line, flowable_x)

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        if fringe_layout.pos_x_in_staff == self.pos_x_in_staff:
            pos = Point(pos.x + fringe_layout.clef, pos.y)
        super().render_complete(pos, flowable_line)

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        super().render_complete(Point(pos.x + fringe_layout.clef, pos.y), flowable_line)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        super().render_complete(Point(pos.x + fringe_layout.clef, pos.y), flowable_line)



from enum import Enum
from typing import Optional


class AccidentalType(Enum):

    """A logical accidental descriptor covering standard 12-EDO western accidentals.

    These are used both in abstract :obj:`.Pitch` descriptions and in concrete
    graphical :obj:`.Accidental` objects.

    Each enum's string value is its corresponding canonical SMuFL glyph name.

    Not all glyphs are mandatory in SMuFL, but all are available in Bravura.
    """

    FLAT = "accidentalFlat"

    NATURAL = "accidentalNatural"

    SHARP = "accidentalSharp"

    DOUBLE_FLAT = "accidentalDoubleFlat"

    DOUBLE_SHARP = "accidentalDoubleSharp"

    NATURAL_FLAT = "accidentalNaturalFlat"

    NATURAL_SHARP = "accidentalNaturalSharp"

    FLAT_SMALL = "accidentalFlatSmall"

    NATURAL_SMALL = "accidentalNaturalSmall"

    SHARP_SMALL = "accidentalSharpSmall"

    FLAT_PARENS = "accidentalFlatParens"

    NATURAL_PARENS = "accidentalNaturalParens"

    SHARP_PARENS = "accidentalSharpParens"

    DOUBLE_SHARP_PARENS = "accidentalDoubleSharpParens"

    DOUBLE_FLAT_PARENS = "accidentalDoubleFlatParens"

    @property
    def pitch_class_offset(self) -> Optional[int]:
        """Return an integer pitch class offset if applicable.

        This will return ``None`` for all accidental types except standard 12-EDO ones.
        """
        return _offset_map.get(self, None)


_offset_map = {
    AccidentalType.FLAT: -1,
    AccidentalType.NATURAL: 0,
    AccidentalType.SHARP: 1,
    AccidentalType.DOUBLE_FLAT: -2,
    AccidentalType.DOUBLE_SHARP: 2,
    AccidentalType.NATURAL_FLAT: 0,
    AccidentalType.NATURAL_SHARP: 1,
    AccidentalType.FLAT_SMALL: -1,
    AccidentalType.NATURAL_SMALL: 0,
    AccidentalType.SHARP_SMALL: 1,
    AccidentalType.FLAT_PARENS: -1,
    AccidentalType.NATURAL_PARENS: 0,
    AccidentalType.SHARP_PARENS: 1,
    AccidentalType.DOUBLE_SHARP_PARENS: 2,
    AccidentalType.DOUBLE_FLAT_PARENS: -2,
}



from dataclasses import dataclass, field
from typing import Optional

from PyQt5 import QtGui

from neoscore.core import neoscore
from neoscore.core.rect import Rect
from neoscore.core.units import Unit
from neoscore.interface.qt.converters import qt_rect_to_rect


@dataclass(frozen=True)
class FontInterface:

    """An interface for fonts, exposing some commonly used metadata."""

    family_name: str
    size: Unit
    weight: Optional[int]
    italic: bool

    ascent: Unit = field(init=False)
    """The ascent of the font.

    The ascent is the vertical distance between the font baseline and
    the highest any font characters reach.
    """

    descent: Unit = field(init=False)
    """The descent of the font.

    The ascent is the vertical distance between the font baseline and
    the lowest any font characters reach.
    """

    x_height: Unit = field(init=False)
    """The x-height for the font.

    This is generally similar, if not identical, to the em size.
    """

    qt_object: QtGui.QFont = field(init=False)

    _qt_font_info_object: QtGui.QFontInfo = field(init=False)
    _qt_font_metrics_object: QtGui.QFontMetricsF = field(init=False)

    def __post_init__(self):
        super().__setattr__(
            "qt_object",
            QtGui.QFont(
                self.family_name,
                # Float font sizes can't be set in QFont's constructor,
                # so set it to -1 (system default) here and set actual
                # size below with setPointSizeF
                -1,
                self.weight if self.weight is not None else -1,
                self.italic,
            ),
        )
        self.qt_object.setPixelSize(round(self.size.base_value))
        super().__setattr__("_qt_font_info_object", QtGui.QFontInfo(self.qt_object))
        super().__setattr__(
            "_qt_font_metrics_object",
            QtGui.QFontMetricsF(self.qt_object, neoscore.app_interface.view),
        )
        super().__setattr__("ascent", Unit(self._qt_font_metrics_object.ascent()))
        super().__setattr__("descent", Unit(self._qt_font_metrics_object.descent()))

        super().__setattr__("x_height", Unit(self._qt_font_metrics_object.xHeight()))

    def bounding_rect_of(self, text: str) -> Rect:
        """Calculate the tight bounding rectangle around a string in this font."""
        # Qt warns that tightBoundingRect is very slow on Windows.
        # Maybe horizontalAdvance can be used instead?
        return qt_rect_to_rect(self._qt_font_metrics_object.tightBoundingRect(text))



from __future__ import annotations

from typing import Optional

from neoscore.core import neoscore
from neoscore.core.font import Font
from neoscore.core.layout_controllers import NewLine
from neoscore.core.point import ORIGIN, Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO, Unit
from neoscore.interface.rich_text_interface import RichTextInterface


class RichText(PositionedObject):

    """A graphical rich text object.

    Unlike its simpler counterpart in ``Text``, this supports rich
    formatting with HTML input. As neoscore is currently backed by Qt,
    this input supports Qt's HTML subset documented at
    https://doc.qt.io/qt-5/richtext-html-subset.html
    """

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        html_text: str,
        width: Optional[Unit] = None,
        font: Optional[Font] = None,
        scale: float = 1,
        rotation: float = 0,
        transform_origin: PointDef = ORIGIN,
    ):
        """
        Args:
            pos: Position relative to the parent
            parent: The parent (core-level) object or None
            html_text: The text to be displayed, written in HTML.
            width: The maximum width for the displayed text. If omitted,
                lines will only break when explicitly called for by the HTML.
            font: The default font to display the text in.
            scale: A scaling factor relative to the font size.
            rotation: Rotation angle in degrees.
            transform_origin: The origin point for rotation and scaling transforms
        """
        super().__init__(pos, parent)
        if font:
            self._font = font
        else:
            self._font = neoscore.default_font
        self._html_text = html_text
        self._width = width
        self._scale = scale
        self._rotation = rotation
        self.transform_origin = transform_origin

    @property
    def breakable_length(self) -> Unit:
        """The breakable length of the object.

        This is always 0, meaning RichText objects cannot be broken
        across Flowable lines.
        """
        return ZERO

    @property
    def html_text(self) -> str:
        """The text to be displayed, written in HTML."""
        return self._html_text

    @html_text.setter
    def html_text(self, value: str):
        self._html_text = value

    @property
    def width(self) -> Optional[Unit]:
        """The maximum width for the displayed text.

        If omitted, lines will only break when explicitly called for
        by the HTML.
        """
        return self._width

    @width.setter
    def width(self, value: Optional[Unit]):
        self._width = value

    @property
    def font(self) -> Font:
        """The default font used to display the text.

        HTML markup can override this to different styles and famililes.
        """
        return self._font

    @font.setter
    def font(self, value: Font):
        self._font = value

    # Since RichText isn't breakable (for now?), we only need to
    # implement complete rendering
    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        interface = RichTextInterface(
            pos,
            None if flowable_line else self.parent.interface_for_children,
            self.scale,
            self.rotation,
            self.transform_origin,
            self.html_text,
            self.font.interface,
            self.width,
        )
        interface.render()
        self.interfaces.append(interface)



from dataclasses import dataclass
from typing import Dict, Optional

from neoscore.core.point import Point
from neoscore.core.rect import Rect
from neoscore.core.units import Unit


@dataclass
class GlyphInfo:
    """SMuFL glyph metadata."""

    canonical_name: str
    """The canonical glyph name, e.g. 'gClefFlat1Below'"""

    codepoint: str
    """The glyph as a unicode string, e.g. ``'\\uF55D'``"""

    description: str
    """Short description of glyph, e.g. 'G clef, flat 1 below'"""

    bounding_rect: Optional[Rect]
    """Glyph bounding rect from SMuFL metadata.

    Note that this is not always very accurate. If accurate measurements are needed,
    it's better to query the font with :obj:`.Font.bounding_rect_of`.
    """

    advance_width: Optional[Unit]
    """The typographic advance distance specified for the glyph."""

    anchors: Optional[Dict[str, Point]]
    """A collection of anchor points provided by SMuFL metadata.

    See `SMuFL's docs on anchors
    <https://w3c.github.io/smufl/latest/specification/glyphswithanchors.html>`_. Note
    that, while SMuFL provides anchors in a vertically inverted coordinate system, they
    are automatically translated to neoscore's coordinates here.
    """



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple, Union

from typing_extensions import TypeAlias

from neoscore.core.point import Point
from neoscore.core.units import Unit


@dataclass(frozen=True)
class Rect:
    """A rectangle data class.

    The ``x`` and ``y`` coordinates represent the starting position,
    typically the top left corner. ``width`` and ``height`` extend
    rightward and downward.

    The only math operation supported for rects is scalar
    multiplication, where each field is multiplied by a given number.
    """

    x: Unit
    y: Unit
    width: Unit
    height: Unit

    @classmethod
    def from_def(cls, rect_def: RectDef) -> Rect:
        if isinstance(rect_def, Rect):
            return rect_def
        return cls(*rect_def)

    def offset(self, offset: Point) -> Rect:
        """Translate a rect by a point."""
        return Rect(self.x + offset.x, self.y + offset.y, self.width, self.height)

    def merge(self, other: Rect) -> Rect:
        """Find the rect encompassing both ``self`` and ``other``.

        Note: This assumes ``width`` and ``height`` in both rects are positive.
        """
        x = min(self.x, other.x)
        y = min(self.y, other.y)
        right_edge = max(self.x + self.width, other.x + other.width)
        bottom_edge = max(self.y + self.height, other.y + other.height)
        return Rect(x, y, right_edge - x, bottom_edge - y)

    def __mul__(self, other: float) -> Rect:
        return Rect(
            self.x * other, self.y * other, self.width * other, self.height * other
        )


RectDef: TypeAlias = Union[Rect, Tuple[Unit, Unit, Unit, Unit]]
"""A ``Rect`` or an init arg tuple for one."""



from __future__ import annotations

from typing import Any, Optional, Union

from typing_extensions import TypeAlias

from neoscore.core.brush_pattern import BrushPattern
from neoscore.core.color import Color, ColorDef
from neoscore.interface.brush_interface import BrushInterface


class Brush:

    """A brush describing how shapes are filled in."""

    _default_color = Color("#000000")

    def __init__(
        self,
        color: Optional[ColorDef] = None,
        pattern: BrushPattern = BrushPattern.SOLID,
    ):
        """
        Args:
            color: The brush color. Defaults to black unless changed globally by
                ``neoscore.set_default_color``.
            pattern: The brush fill pattern.
        """
        if color is None:
            self._color = Brush._default_color
        else:
            self._color = Color.from_def(color)
        self._pattern = pattern
        self._regenerate_interface()

    @classmethod
    def from_existing(
        cls,
        brush: Brush,
        color: Optional[ColorDef] = None,
        pattern: Optional[BrushPattern] = None,
    ) -> Brush:
        """Derive a Brush from another, overriding any provided fields."""
        return cls(color or brush.color, pattern or brush.pattern)

    @classmethod
    def from_def(cls, brush_def: BrushDef) -> Brush:
        if isinstance(brush_def, Brush):
            return brush_def
        else:
            return Brush(brush_def)

    @classmethod
    def no_brush(cls) -> Brush:
        """Create a non-drawing brush."""
        return Brush(pattern=BrushPattern.INVISIBLE)

    def _regenerate_interface(self):
        self._interface = BrushInterface(self.color, self.pattern)

    @property
    def color(self) -> Color:
        """The color for the brush"""
        return self._color

    @color.setter
    def color(self, value: Color):
        self._color = value
        self._regenerate_interface()

    @property
    def pattern(self) -> BrushPattern:
        """The fill pattern."""
        return self._pattern

    @pattern.setter
    def pattern(self, value: BrushPattern):
        self._pattern = value
        self._regenerate_interface()

    @property
    def interface(self) -> BrushInterface:
        return self._interface

    def __eq__(self, other: Any) -> bool:
        """Brushes are compared by their attributes"""
        return (
            isinstance(other, Brush)
            and self.color == other.color
            and self.pattern == other.pattern
        )


BrushDef: TypeAlias = Union[Brush, str]
"""A ``Brush`` or a color hex string to be passed to an otherwise default ``Brush``."""



from neoscore.core.units import Unit
from neoscore.western.clef import Clef
from neoscore.western.clef_type import ClefTypeDef
from neoscore.western.staff import Staff


class InvisibleClef(Clef):

    """A non-printing clef.

    This is useful in contexts where staves are used with implied clefs.
    """

    def __init__(self, pos_x: Unit, staff: Staff, clef_type: ClefTypeDef):
        # Hackily create a superclass Clef, then just overwrite its text
        # this is slightly inefficient but probably fine
        super().__init__(pos_x, staff, clef_type)
        self.text = ""



from __future__ import annotations

from typing import TYPE_CHECKING, Type, cast

from neoscore.core.exceptions import NoAncestorWithMusicFontError
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import Unit

if TYPE_CHECKING:
    from neoscore.core.music_font import MusicFont


class HasMusicFont:
    """A mixin for classes which have an associated staff measurement.

    This is useful for integration with SMuFL fonts, where glyphs and their metadata are
    measured relative to a staff space.

    A naive and slow implementation of the :obj:`music_font <.HasMusicFont.music_font>`
    property is provided which returns the nearest music found in the object's
    ancestors. Subclasses will often want to override this.
    """

    @property
    def music_font(self) -> MusicFont:
        """The music font used by this object."""
        return HasMusicFont.find_music_font(cast(PositionedObject, self).parent)

    @property
    def unit(self) -> Type[Unit]:
        """A unit type where ``unit(1)`` is the size of a staff space in the font."""
        return self.music_font.unit

    @staticmethod
    def find_music_font(obj: PositionedObject) -> MusicFont:
        """Return the music font from ``obj`` or its nearest ancestor with one.

        This checks ``obj`` and then its ancestors for whether they have a
        ``music_font`` attribute, returning the first value found. If no font is found,
        this raises a :obj:`.NoAncestorWithMusicFontError`.
        """
        if hasattr(obj, "music_font"):
            return cast(HasMusicFont, obj).music_font
        lookup_result = obj.first_ancestor_with_attr("music_font")
        if lookup_result is None:
            raise NoAncestorWithMusicFontError()
        return cast(HasMusicFont, lookup_result).music_font



class BreakOpportunity:
    """A :obj:`.PositionedObject` mixin which indicates a break can be performed here.

    Some notational constructs are well suited to be placed at line
    or page breaks, for example bar lines or grand pauses. Such classes
    should subclass this mixin so the flowable container can know this.

    This mixin is only useful when used inside a :obj:`.Flowable`.
    """

    # Type sentinel for faster type lookups
    _neoscore_break_opportunity_type_marker = True



from __future__ import annotations

import math
from typing import cast

from neoscore.core.math_helpers import point_angle
from neoscore.core.point import Point, PointDef
from neoscore.core.positioned_object import PositionedObject, render_cached_property
from neoscore.core.spanner import Spanner
from neoscore.core.units import Unit


class Spanner2D(Spanner):

    """A 2-dimensional :obj:`.Spanner`."""

    def __init__(self, end_pos: PointDef, end_parent: PositionedObject):
        """
        Args:
            end_pos: The end point.
            end_parent: The parent for the ending position.
        """
        end_pos = Point.from_def(end_pos)
        super().__init__(end_pos.x, end_parent)
        self._end_y = end_pos.y

    @property
    def end_y(self) -> Unit:
        """The y position of the endpoint as specified."""
        return self._end_y

    @end_y.setter
    def end_y(self, value: Unit):
        self._end_y = value

    @property
    def end_pos(self) -> Point:
        return Point(self._end_x, self._end_y)

    @end_pos.setter
    def end_pos(self, value: PointDef):
        value = Point.from_def(value)
        self._end_x = value.x
        self._end_y = value.y

    @render_cached_property
    def spanner_2d_length(self) -> Unit:
        """The 2d length of the spanner.

        This takes into account both the x and y-axis. For only the horizontal length,
        use :obj:`.Spanner.spanner_x_length`.
        """
        return cast(PositionedObject, self).distance_to(self.end_parent, self.end_pos)

    @render_cached_property
    def angle(self) -> float:
        """The angle from the start to end point in degrees.

        The angle goes from the positive X axis to the end point. Positive angles go
        clockwise.
        """
        return math.degrees(point_angle(self._relative_end_pos))

    @render_cached_property
    def _relative_end_pos(self) -> Point:
        return cast(PositionedObject, self).map_to(self.end_parent) + self.end_pos



from __future__ import annotations

from typing import Optional

from neoscore.core.brush import BrushDef
from neoscore.core.directions import DirectionY
from neoscore.core.math_helpers import interpolate
from neoscore.core.music_font import MusicFont
from neoscore.core.music_path import MusicPath
from neoscore.core.pen import PenDef
from neoscore.core.point import ORIGIN, Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO, Unit


class AbstractSlur(MusicPath):
    """An abstract superclass for slur and tie.

    This is not meant to be used directly."""

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        direction: DirectionY = DirectionY.UP,
        height: Optional[Unit] = None,
        arch_length: Optional[Unit] = None,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        MusicPath.__init__(self, pos, parent, font, brush, pen)
        self.direction = direction
        self.height = height
        self.arch_length = arch_length
        # Load relevant engraving defaults from music font
        self.midpoint_thickness = self.music_font.engraving_defaults[
            "slurMidpointThickness"
        ]
        self.endpoint_thickness = self.music_font.engraving_defaults[
            "slurEndpointThickness"
        ]

    def draw_slur(self, end_pos, end_parent):
        """Draw the slur shape.

        This should generally not be called directly.
        """
        # Work out parameters
        length = self.distance_to(end_parent, end_pos)
        abs_height = self.height if self.height else self._derive_height(length)
        arch_length = (
            self.arch_length if self.arch_length else self._derive_arch_length(length)
        )
        mid_height = abs_height * self.direction.value
        mid_upper_height = mid_height + (self.midpoint_thickness * self.direction.value)
        end_height = self.endpoint_thickness * self.direction.value
        # Draw upper curve part
        self.move_to(ZERO, end_height, self)
        control_1 = Point(arch_length, mid_upper_height)
        control_1_parent = self
        control_2 = Point(
            self.end_pos.x - arch_length, self.end_pos.y + mid_upper_height
        )
        control_2_parent = self.end_parent
        end = Point(self.end_pos.x, self.end_pos.y + end_height)
        end_parent = self.end_parent
        self.cubic_to(
            control_1.x,
            control_1.y,
            control_2.x,
            control_2.y,
            end.x,
            end.y,
            control_1_parent,
            control_2_parent,
            end_parent,
        )
        # Draw right-side end
        self.line_to(self.end_pos.x, self.end_pos.y, self.end_parent)
        # Draw lower curve part
        control_1 = Point(self.end_pos.x - arch_length, self.end_pos.y + mid_height)
        control_1_parent = self.end_parent
        control_2 = Point(arch_length, mid_height)
        control_2_parent = self
        end = ORIGIN
        end_parent = self
        self.cubic_to(
            control_1.x,
            control_1.y,
            control_2.x,
            control_2.y,
            end.x,
            end.y,
            control_1_parent,
            control_2_parent,
            end_parent,
        )

    def _derive_height(self, length):
        # length = self.spanner_2d_length
        unit = self.music_font.unit
        max_length_considered = unit(10)
        min_length_considered = unit(4)
        max_height = unit(2)
        min_height = unit(0.75)
        if length > max_length_considered:
            return max_height
        if length < min_length_considered:
            return min_height
        return interpolate(
            Point(min_length_considered, min_height),
            Point(max_length_considered, max_height),
            length,
        )

    def _derive_arch_length(self, length):
        # spanner_length = self.spanner_2d_length
        unit = self.music_font.unit
        max_spanner_length_considered = unit(10)
        min_spanner_length_considered = unit(4)
        max_arch_length = unit(1)
        min_arch_length = unit(0)
        if length > max_spanner_length_considered:
            return max_arch_length
        if length < min_spanner_length_considered:
            return min_arch_length
        return interpolate(
            Point(min_spanner_length_considered, min_arch_length),
            Point(max_spanner_length_considered, max_arch_length),
            length,
        )



"""General math helper tools."""

import math
from typing import Tuple

from neoscore.core.point import Point
from neoscore.core.units import Unit


def interpolate(line_start: Point, line_end: Point, x: Unit) -> Unit:
    """Calculate the value of a line defined by two points at a given x pos."""
    slope = (line_end.y - line_start.y) / (line_end.x - line_start.x)
    y_intercept = line_start.y - (line_start.x * slope)
    return (x * slope) + y_intercept


def is_power_of_2(value: int) -> bool:
    """Check if a number is a power of 2."""
    return (value & (value - 1) == 0) and value != 0


def point_angle(point: Point) -> float:
    """Calculate the angle from the positive X axis to a point in radians.

    The returned angle goes positive clockwise and is between pi and -pi.

    Combine with the Python stdlib function ``math.degrees()`` to get
    the angle in degrees.
    """
    return math.atan2(point.y.base_value, point.x.base_value)


def dist(p1: Tuple[float, float], p2: Tuple[float, float]) -> float:
    """Find the Euclidean distance between two 2D float tuples.

    In the future this function and signature may be modified to support neoscore
    unit-points as well.
    """
    return math.sqrt(((p2[1] - p1[1]) ** 2) + ((p2[0] - p1[0]) ** 2))



from __future__ import annotations

from typing import Optional

from neoscore.core import neoscore
from neoscore.core.brush import Brush, BrushDef
from neoscore.core.font import Font
from neoscore.core.layout_controllers import NewLine
from neoscore.core.painted_object import PaintedObject
from neoscore.core.pen import Pen, PenDef
from neoscore.core.point import ORIGIN, Point, PointDef
from neoscore.core.positioned_object import PositionedObject, render_cached_property
from neoscore.core.rect import Rect
from neoscore.core.text_alignment import AlignmentX, AlignmentY
from neoscore.core.units import ZERO, Unit
from neoscore.interface.text_interface import TextInterface


class Text(PaintedObject):

    """A graphical text object."""

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        text: str,
        font: Optional[Font] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        scale: float = 1,
        rotation: float = 0,
        background_brush: Optional[BrushDef] = None,
        breakable: bool = True,
        alignment_x: AlignmentX = AlignmentX.LEFT,
        alignment_y: AlignmentY = AlignmentY.BASELINE,
        transform_origin: PointDef = ORIGIN,
    ):
        """
        Args:
            pos: Position relative to the parent
            parent: The parent object. Defaults to the document's first page.
            text: The text to be displayed
            font: The font to display the text in.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
            scale: A scaling factor relative to the font size.
            rotation: Angle in degrees. Note that breakable rotated text is
                not currently supported.
            background_brush: Optional brush used to paint the text's bounding rect
                behind it.
            breakable: Whether this object should break across lines in
                :obj:`.Flowable` containers.
            alignment_x: The text's horizontal alignment relative to ``pos``.
                Note that text which is not ``LEFT`` aligned does not currently display
                correctly when breaking across flowable lines.
            alignment_y: The text's vertical alignment relative to ``pos``.
            transform_origin: The origin point for rotation and scaling transforms
        """
        super().__init__(pos, parent, brush, pen or Pen.no_pen())
        if font:
            self._font = font
        else:
            self._font = neoscore.default_font
        self._text = text
        self._scale = scale
        self._rotation = rotation
        self.background_brush = background_brush
        self._breakable = breakable
        self._alignment_x = alignment_x
        self._alignment_y = alignment_y
        self.transform_origin = transform_origin

    @property
    def breakable_length(self) -> Unit:
        """The breakable length of the object.

        This is used to determine how and where rendering cuts should be made.

        This is derived from other properties and cannot be set directly.
        """
        if self.breakable:
            return self.bounding_rect.width + self._alignment_offset.x
        else:
            return ZERO

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, value: str):
        self._text = value

    @property
    def font(self) -> Font:
        return self._font

    @font.setter
    def font(self, value: Font):
        self._font = value

    @property
    def background_brush(self) -> Optional[Brush]:
        """The brush to paint over the background with."""
        return self._background_brush

    @background_brush.setter
    def background_brush(self, value: Optional[BrushDef]):
        if value:
            self._background_brush = Brush.from_def(value)
        else:
            self._background_brush = None

    @property
    def breakable(self) -> bool:
        """Whether this object should be broken across flowable lines."""
        return self._breakable

    @breakable.setter
    def breakable(self, value: bool):
        self._breakable = value

    @property
    def alignment_x(self) -> AlignmentX:
        """The text's horizontal alignment relative to ``pos``.

        Note that text which is not ``LEFT`` aligned does not currently display
        correctly when breaking across flowable lines.
        """
        return self._alignment_x

    @alignment_x.setter
    def alignment_x(self, value: AlignmentX):
        self._alignment_x = value

    @property
    def alignment_y(self) -> AlignmentY:
        """The text's vertical alignment relative to ``pos``."""
        return self._alignment_y

    @alignment_y.setter
    def alignment_y(self, value: AlignmentY):
        self._alignment_y = value

    @render_cached_property
    def _alignment_offset(self) -> Point:
        if (
            self.alignment_x == AlignmentX.LEFT
            and self.alignment_y == AlignmentY.BASELINE
        ):
            return ORIGIN
        x = ZERO
        y = ZERO
        bounding_rect = self._raw_scaled_bounding_rect
        if self.alignment_x == AlignmentX.CENTER:
            x = (bounding_rect.width / -2) - bounding_rect.x
        elif self.alignment_x == AlignmentX.RIGHT:
            x = -bounding_rect.width - bounding_rect.x
        if self.alignment_y == AlignmentY.CENTER:
            y = (bounding_rect.height / -2) - bounding_rect.y
        return Point(x, y)

    @render_cached_property
    def _raw_scaled_bounding_rect(self) -> Rect:
        """The text bounding rect without centering adjustment"""
        return self.font.bounding_rect_of(self.text) * self.scale

    @render_cached_property
    def bounding_rect(self) -> Rect:
        """The bounding rect for this text positioned relative to ``pos``.

        The rect ``(x, y)`` position is relative to the object's position.

        Note that this currently accounts for scaling, but not rotation.
        """
        raw_rect = self._raw_scaled_bounding_rect
        alignment_offset = self._alignment_offset
        return Rect(
            raw_rect.x + alignment_offset.x,
            raw_rect.y + alignment_offset.y,
            raw_rect.width,
            raw_rect.height,
        )

    def _render_slice(
        self,
        pos: Point,
        inside_flowable: bool,
        clip_start_x: Optional[Unit] = None,
        clip_width: Optional[Unit] = None,
    ):
        """Render a horizontal slice of a text object.

        Args:
            pos: The doc-space position of the slice beginning
                (at the top-left corner of the slice)
            clip_start_x: The starting local x position in of the slice
            clip_width: The horizontal length of the slice to
                be rendered
            inside_flowable: Whether this is being rendered in a flowable.
                This affects the treatment of the interface position and parent.
        """
        slice_interface = TextInterface(
            pos + self._alignment_offset,
            None if inside_flowable else self.parent.interface_for_children,
            self.scale,
            self.rotation,
            self.transform_origin,
            self.brush.interface,
            self.pen.interface,
            self.text,
            self.font.interface,
            self.background_brush.interface if self.background_brush else None,
            clip_start_x,
            clip_width,
        )
        slice_interface.render()
        self.interfaces.append(slice_interface)

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        inside_flowable = bool(flowable_line)
        self._render_slice(pos, inside_flowable, None, None)

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        slice_length = flowable_line.length - (flowable_x - flowable_line.flowable_x)
        self._render_slice(pos, True, ZERO, slice_length)

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        self._render_slice(pos, True, object_x, flowable_line.length)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        self._render_slice(pos, True, object_x, None)



from collections.abc import Callable
from typing import Optional

from neoscore.core.brush import Brush
from neoscore.core.page_supplier import PageOverlayFunc, PageSupplier
from neoscore.core.paper import Paper
from neoscore.core.point import Point
from neoscore.core.units import ZERO, Mm

_PAGE_DISPLAY_GAP = Mm(50)


class Document:

    """The document root object.

    This object should not be created directly by users - it is instantiated
    by :obj:`.neoscore.setup`, which creates a global instance of this class which
    can be then accessed as :obj:`.neoscore.document`.
    """

    def __init__(self, paper: Paper, overlay_func: Optional[PageOverlayFunc] = None):
        """
        Args:
            paper: The paper to use in the document.
            overlay_func: An optional function to run on each generated page.
        """
        self._paper = paper
        self._pages = PageSupplier(self, overlay_func)

    @property
    def paper(self) -> Paper:
        """The paper type of the document"""
        return self._paper

    @paper.setter
    def paper(self, value):
        self._paper = value

    @property
    def pages(self) -> PageSupplier:
        """The document's pages.

        Pages are created on-demand by accessing this property.

        This property can be treated like a managed list:

            >>> from neoscore.core import neoscore; neoscore.setup()
            >>> len(neoscore.document.pages)             # No pages exist yet
            0
            >>> first_page = neoscore.document.pages[0]  # Get the first page
            >>> len(neoscore.document.pages)             # One page now exists
            1
            >>> sixth_page = neoscore.document.pages[5]  # Get the sixth page
            >>> len(neoscore.document.pages)             # 5 new pages are created
            6
            >>> # Pages can be accessed by negative indexing too
            >>> assert(first_page == neoscore.document.pages[-6])
            >>> assert(sixth_page == neoscore.document.pages[-1])
            >>> neoscore.shutdown()

        For more information on this object, see :obj:`.PageSupplier`.
        """
        return self._pages

    def _run_on_all_descendants(self, func: Callable):
        for page in self.pages:
            for obj in page.descendants:
                func(obj)

    def render(self, display_page_geometry: bool, background_brush: Brush):
        """Render all items in the document.

        This should not be called directly.

        Args:
            display_page_geometry: Whether to include a preview of page geometry.
            background_brush: The brush used to draw the scene background.
        """
        self._run_on_all_descendants(lambda g: g.pre_render_hook())
        if display_page_geometry:
            for page in self.pages:
                page.create_geometry_preview(background_brush)
        for page in self.pages:
            page.render()
        self._run_on_all_descendants(lambda g: g.post_render_hook())

    def page_origin(self, index: int) -> Point:
        """Find the origin point of a given page number.

        The origin is the top left corner of the live page area.

        Args:
            index: The 0-based index of the page to locate.
        """
        page_x = (self.paper.width + _PAGE_DISPLAY_GAP) * index
        return Point(page_x, ZERO)



"""Tables of glyphs for common notehead styles
This module includes prebuilt ``NoteheadTable``\ s for many notehead
flavors included in SMuFL. Custom tables using arbitrary SMuFL glyphs
may also be created by simply instantiating a ``NoteheadTable`` and
passing it wherever needed.

Note that many of the more obscure glyph tables only have a solid and
empty variant, for example ``MOON``. In these cases, the empty glyph is
used for all except the ``short`` key, for which the solid glyph is
used.
"""

from dataclasses import dataclass
from typing import List

from neoscore.western.duration_display import BaseDuration

# NOTE: When creating a new table, be sure to add it to ALL_TABLES at
# the end of this file.


@dataclass(frozen=True)
class NoteheadTable:
    """A mapping from duration classes to canonical SMuFL glyph names."""

    double_whole: str
    whole: str
    half: str
    short: str
    """Quarter notes and shorter"""

    def lookup_duration(self, base_division: BaseDuration) -> str:
        """Given a base division, return the required glyph to represent it.

        Assumes ``base_division`` is 0 or a power of 2.
        """
        if base_division >= 4:
            return self.short
        elif base_division == 2:
            return self.half
        elif base_division == 1:
            return self.whole
        else:
            return self.double_whole


STANDARD = NoteheadTable(
    "noteheadDoubleWhole", "noteheadWhole", "noteheadHalf", "noteheadBlack"
)
"""Standard western notation noteheads"""

STANDARD_WITH_PARENTHESES = NoteheadTable(
    "noteheadDoubleWholeParens",
    "noteheadWholeParens",
    "noteheadHalfParens",
    "noteheadBlackParens",
)
"""Standard noteheads enclosed in parentheses.

Not available in all fonts.
"""

STANDARD_SMALL = NoteheadTable(
    "noteheadDoubleWholeSmall",
    "noteheadWholeSmall",
    "noteheadHalfSmall",
    "noteheadBlackSmall",
)
"""Standard noteheads at a smaller size.

Not available in all fonts.
"""

STANDARD_OVERSIZED = NoteheadTable(
    "noteheadDoubleWholeOversized",
    "noteheadWholeOversized",
    "noteheadHalfOversized",
    "noteheadBlackOversized",
)
"""Standard noteheads at a larger size.

Not available in all fonts.
"""

X = NoteheadTable(
    "noteheadXDoubleWhole", "noteheadXWhole", "noteheadXHalf", "noteheadXBlack"
)
""""""

SLASH = NoteheadTable(
    "noteheadSlashWhiteDoubleWhole",
    "noteheadSlashWhiteWhole",
    "noteheadSlashWhiteHalf",
    "noteheadSlashVerticalEnds",
)
"""Percussion-style slash noteheads.

This uses SMuFL's vertical-ends-style short notehead.
"""

MUTED_SLASH = NoteheadTable(
    "noteheadSlashWhiteDoubleWhole",
    "noteheadSlashWhiteMuted",
    "noteheadSlashWhiteMuted",
    "noteheadSlashVerticalEndsMuted",
)
"""Percussion-style slash noteheads with secondary slashes indicating mutes.

Because SMuFL provides no muted variant for double whole notes, this
uses the regular slash double whole notehead.
"""

SLASH_OVERSIZED = NoteheadTable(
    "noteheadSlashWhiteDoubleWholeOversized",
    "noteheadSlashWhiteWholeOversized",
    "noteheadSlashWhiteHalfOversized",
    "noteheadSlashVerticalEndsOversized",
)
"""Like ``SLASH`` but with oversized variants.

Not available in all fonts.
"""

MUTED_SLASH_OVERSIZED = NoteheadTable(
    "noteheadSlashWhiteDoubleWholeOversized",
    "noteheadSlashWhiteMutedOversized",
    "noteheadSlashWhiteMutedOversized",
    "noteheadSlashVerticalEndsMutedOversized",
)
"""Like ``MUTED_SLASH`` but with oversized variants.

Not available in all fonts.
"""

PLUS = NoteheadTable(
    "noteheadPlusDoubleWhole",
    "noteheadPlusWhole",
    "noteheadPlusHalf",
    "noteheadPlusBlack",
)
""""""

CIRCLE_X = NoteheadTable(
    "noteheadCircleXDoubleWhole",
    "noteheadCircleXWhole",
    "noteheadCircleXHalf",
    "noteheadCircleX",
)
""""""

STANDARD_WITH_X = NoteheadTable(
    "noteheadDoubleWholeWithX",
    "noteheadWholeWithX",
    "noteheadHalfWithX",
    "noteheadVoidWithX",
)
""""""

SQUARE = NoteheadTable(
    "noteheadSquareWhite",
    "noteheadSquareWhite",
    "noteheadSquareWhite",
    "noteheadSquareBlack",
)
""""""

TRIANGLE_UP = NoteheadTable(
    "noteheadTriangleUpDoubleWhole",
    "noteheadTriangleUpWhole",
    "noteheadTriangleUpHalf",
    "noteheadTriangleUpBlack",
)
""""""

TRIANGLE_LEFT = NoteheadTable(
    "noteheadTriangleLeftWhite",
    "noteheadTriangleLeftWhite",
    "noteheadTriangleLeftWhite",
    "noteheadTriangleLeftBlack",
)
""""""

TRIANGLE_RIGHT = NoteheadTable(
    "noteheadTriangleRightWhite",
    "noteheadTriangleRightWhite",
    "noteheadTriangleRightWhite",
    "noteheadTriangleRightBlack",
)
""""""

TRIANGLE_DOWN = NoteheadTable(
    "noteheadTriangleDownDoubleWhole",
    "noteheadTriangleDownWhole",
    "noteheadTriangleDownHalf",
    "noteheadTriangleDownBlack",
)
""""""

TRIANGLE_UP_RIGHT = NoteheadTable(
    "noteheadTriangleUpRightWhite",
    "noteheadTriangleUpRightWhite",
    "noteheadTriangleUpRightWhite",
    "noteheadTriangleUpRightBlack",
)
""""""

TRIANGLE_ROUND_DOWN = NoteheadTable(
    "noteheadTriangleRoundDownWhite",
    "noteheadTriangleRoundDownWhite",
    "noteheadTriangleRoundDownWhite",
    "noteheadTriangleRoundDownBlack",
)
""""""

MOON = NoteheadTable(
    "noteheadMoonWhite",
    "noteheadMoonWhite",
    "noteheadMoonWhite",
    "noteheadMoonBlack",
)
""""""

PARENTHESIS = NoteheadTable(
    "noteheadParenthesis",
    "noteheadParenthesis",
    "noteheadParenthesis",
    "noteheadParenthesis",
)
"""An empty set of parentheses in place of noteheads. All four glyphs are the same."""

STANDARD_WITH_SLASH_UP = NoteheadTable(
    "noteheadSlashedDoubleWhole1",
    "noteheadSlashedWhole1",
    "noteheadSlashedHalf1",
    "noteheadSlashedBlack1",
)
"""Standard noteheads with slashes going from bottom left to top right."""

STANDARD_WITH_SLASH_DOWN = NoteheadTable(
    "noteheadSlashedDoubleWhole2",
    "noteheadSlashedWhole2",
    "noteheadSlashedHalf2",
    "noteheadSlashedBlack2",
)
"""Standard noteheads with slashes going from top left to bottom right."""

DIAMOND = NoteheadTable(
    "noteheadDiamondDoubleWhole",
    "noteheadDiamondWhole",
    "noteheadDiamondHalf",
    "noteheadDiamondBlack",
)
""""""

DIAMOND_WIDE = NoteheadTable(
    "noteheadDiamondDoubleWhole",
    "noteheadDiamondWhole",
    "noteheadDiamondHalfWide",
    "noteheadDiamondBlackWide",
)
"""Like ``DIAMOND`` but with wider half and short variants"""

DIAMOND_OLD = NoteheadTable(
    "noteheadDiamondDoubleWholeOld",
    "noteheadDiamondWholeOld",
    "noteheadDiamondHalfOld",
    "noteheadDiamondBlackOld",
)
""""""

STANDARD_CIRCLED = NoteheadTable(
    "noteheadCircledDoubleWhole",
    "noteheadCircledWhole",
    "noteheadCircledHalf",
    "noteheadCircledBlack",
)
""""""

STANDARD_LARGE_CIRCLED = NoteheadTable(
    "noteheadCircledDoubleWholeLarge",
    "noteheadCircledWholeLarge",
    "noteheadCircledHalfLarge",
    "noteheadCircledBlackLarge",
)
"""Like ``STANDARD_CIRCLED`` but with a larger circle"""

LARGE_ARROW_UP = NoteheadTable(
    "noteheadLargeArrowUpDoubleWhole",
    "noteheadLargeArrowUpWhole",
    "noteheadLargeArrowUpHalf",
    "noteheadLargeArrowUpBlack",
)
""""""

LARGE_ARROW_DOWN = NoteheadTable(
    "noteheadLargeArrowDownDoubleWhole",
    "noteheadLargeArrowDownWhole",
    "noteheadLargeArrowDownHalf",
    "noteheadLargeArrowDownBlack",
)
""""""

CLUSTER_SQUARE = NoteheadTable(
    "noteheadClusterSquareWhite",
    "noteheadClusterSquareWhite",
    "noteheadClusterSquareWhite",
    "noteheadClusterSquareBlack",
)
"""Large tone clusters in square shapes

This uses the same glyph for all durations except ``short``.
"""

CLUSTER_ROUND = NoteheadTable(
    "noteheadClusterRoundWhite",
    "noteheadClusterRoundWhite",
    "noteheadClusterRoundWhite",
    "noteheadClusterRoundBlack",
)
"""Large tone clusters in round shapes

This uses the same glyph for all durations except ``short``.
"""

CLUSTER_SECOND = NoteheadTable(
    "noteheadClusterDoubleWhole2nd",
    "noteheadClusterWhole2nd",
    "noteheadClusterHalf2nd",
    "noteheadClusterQuarter2nd",
)
"""Tone cluster glyphs spanning a second"""

CLUSTER_THIRD = NoteheadTable(
    "noteheadClusterDoubleWhole3rd",
    "noteheadClusterWhole3rd",
    "noteheadClusterHalf3rd",
    "noteheadClusterQuarter3rd",
)
"""Tone cluster glyphs spanning a third"""

CLUSTER_SECOND_DIAMOND = NoteheadTable(
    "noteheadDiamondClusterWhite2nd",
    "noteheadDiamondClusterWhite2nd",
    "noteheadDiamondClusterWhite2nd",
    "noteheadDiamondClusterBlack2nd",
)
"""Tone cluster glyphs spanning a second using diamond glyphs

This uses the same glyph for all durations except ``short``.
"""

CLUSTER_THIRD_DIAMOND = NoteheadTable(
    "noteheadDiamondClusterWhite3rd",
    "noteheadDiamondClusterWhite3rd",
    "noteheadDiamondClusterWhite3rd",
    "noteheadDiamondClusterBlack3rd",
)
"""Tone cluster glyphs spanning a second using diamond glyphs

This uses the same glyph for all durations except ``short``.
"""

INVISIBLE = NoteheadTable("", "", "", "")
"""Blank non-printing noteheads"""

ALL_TABLES: List[NoteheadTable] = [
    STANDARD,
    STANDARD_WITH_PARENTHESES,
    STANDARD_SMALL,
    STANDARD_OVERSIZED,
    X,
    SLASH,
    MUTED_SLASH,
    SLASH_OVERSIZED,
    MUTED_SLASH_OVERSIZED,
    PLUS,
    CIRCLE_X,
    STANDARD_WITH_X,
    SQUARE,
    TRIANGLE_UP,
    TRIANGLE_LEFT,
    TRIANGLE_RIGHT,
    TRIANGLE_DOWN,
    TRIANGLE_UP_RIGHT,
    TRIANGLE_ROUND_DOWN,
    MOON,
    PARENTHESIS,
    STANDARD_WITH_SLASH_UP,
    STANDARD_WITH_SLASH_DOWN,
    DIAMOND,
    DIAMOND_WIDE,
    DIAMOND_OLD,
    STANDARD_CIRCLED,
    STANDARD_LARGE_CIRCLED,
    LARGE_ARROW_UP,
    LARGE_ARROW_DOWN,
    CLUSTER_SQUARE,
    CLUSTER_ROUND,
    CLUSTER_SECOND,
    CLUSTER_THIRD,
    CLUSTER_SECOND_DIAMOND,
    CLUSTER_THIRD_DIAMOND,
    INVISIBLE,
]
"""A list of all the notehead tables in this module

:meta hide-value:
"""



from __future__ import annotations

from typing import Optional

from neoscore.core.font import Font
from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.music_char import MusicChar
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicStringDef, MusicText
from neoscore.core.point import ORIGIN, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.raw_music_char import RawMusicChar
from neoscore.core.text import Text
from neoscore.core.units import ZERO


class MetronomeMark(PositionedObject, HasMusicFont):
    """A combined :obj:`MusicText` and :obj:`Text` for use in metronome markings"""

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        music_str: MusicStringDef,
        text_str: str,
        music_font: Optional[MusicFont] = None,
        text_font: Optional[Font] = None,
        spaces_between_music_chars: bool = True,
    ):
        """
        Args:
            pos: Position relative to ``parent``
            parent: If no ``music_font`` is given, this or one of its ancestors must
                implement :obj:`.HasMusicFont`.
            music_str: The :obj:`MusicText` string portion of the mark. Glyphs should
                typically come from `the relevant SMuFL range
                <https://w3c.github.io/smufl/latest/tables/metronome-marks.html>`_.
            text_str: The plain-text string portion of the mark. This will typically
                be of the form "= 123".
            music_font: If provided, this overrides any music font found in the
                ancestor chain.
            text_font: The font for the plain text portion of the mark.
            spaces_between_music_chars: Whether to insert spaces between each specified
                music character. This is needed to correctly space rhythm dots if used.
        """
        PositionedObject.__init__(self, pos, parent)
        self._music_font = music_font or HasMusicFont.find_music_font(parent)
        self._music_text_obj = MusicText(ORIGIN, self, music_str, self._music_font)
        if spaces_between_music_chars and len(self._music_text_obj.music_chars) > 1:
            self._music_text_obj.music_chars = self._insert_spaces(
                self._music_text_obj.music_chars
            )
        music_text_rect = self._music_text_obj.bounding_rect
        music_text_end_x = music_text_rect.x + music_text_rect.width
        # Because text objects can't have leading or trailing spaces, to get a space
        # between these elements we do a terrible hack to find the width of a space and
        # offset accordingly. See https://github.com/DigiScore/neoscore/issues/34
        text_obj = Text((music_text_end_x, ZERO), self, " " + text_str, text_font)
        width_including_space = text_obj.bounding_rect.width
        text_obj.text = text_str
        text_obj.x += width_including_space - text_obj.bounding_rect.width
        self._plain_text_obj = text_obj

    @property
    def music_font(self) -> MusicFont:
        return self._music_font

    @property
    def music_text_obj(self) -> MusicText:
        return self._music_text_obj

    @property
    def plain_text_obj(self) -> Text:
        return self._plain_text_obj

    def _insert_spaces(self, chars: List[MusicChar]) -> List[MusicChar]:
        result = []
        blank = RawMusicChar(self.music_font, " ")
        for i, char in enumerate(chars):
            result.append(char)
            if i < len(chars) - 1:
                result.append(blank)
        return result



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Tuple, Union

from backports.cached_property import cached_property
from typing_extensions import TypeAlias

from neoscore.core.glyph_info import GlyphInfo
from neoscore.core.music_font import MusicFont
from neoscore.core.rect import Rect


@dataclass(frozen=True)
class MusicChar:
    """A SMuFL music character.

    Note that measurements in the contained metadata may have slight errors and should
    be used with caution when high precision is needed.
    """

    font: MusicFont
    """The font used to derive SMuFL spec information about this glyph."""

    glyph_name: str
    """The canonical SMuFL name of the glyph"""

    alternate_number: Optional[int] = None
    """An SMuFL alternate glyph code, if applicable."""

    @cached_property
    def glyph_info(self) -> GlyphInfo:
        """SMuFL metadata for this character"""

        return self.font.glyph_info(self.glyph_name, self.alternate_number)

    @property
    def codepoint(self) -> str:
        """The glyph's SMuFL codepoint.

        This is a convenience property for ``glyph_info.codepoint``
        """
        return self.glyph_info.codepoint

    @property
    def bounding_rect(self) -> Rect:
        """The glyph bounding box."""
        return self.glyph_info.bounding_rect


MusicCharDef: TypeAlias = Union[MusicChar, str, Tuple[str, int]]
"""Shorthand for a MusicChar.

Bare ``str`` values should be glyph names, while tuples should be of the form
(``glyph_name, alternate_number``)
"""



from __future__ import annotations

import pathlib
from typing import Optional

from neoscore.core.layout_controllers import NewLine
from neoscore.core.point import ORIGIN, Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO, Unit
from neoscore.interface.image_interface import ImageInterface


class Image(PositionedObject):
    """An image which can be placed in the document.

    Supported image extensions/formats include: BMP, GIF, JPG, JPEG,
    PNG, PBM, PGM, PPM, XBM, XPM, and SVG.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        file_path: str | pathlib.Path,
        scale: float = 1,
        rotation: float = 0,
        transform_origin: PointDef = ORIGIN,
        opacity: float = 1,
    ):
        """
        Args:
            pos: Position relative to the parent
            parent: The parent object or None
            file_path: Path to an image file to be used
            scale: A scaling factor applied to the image.
            rotation: Rotation angle in degrees.
            transform_origin: The origin point for rotation and scaling transforms
            opacity: The image's opacity, where 1 is fully opaque and 0 is invisible.
        """
        super().__init__(pos, parent)
        self._scale = scale
        self._rotation = rotation
        self.transform_origin = transform_origin
        self._opacity = opacity
        self.file_path = file_path

    @property
    def file_path(self) -> pathlib.Path:
        """Path to an image file to be used"""
        return self._file_path

    @file_path.setter
    def file_path(self, value: str | pathlib.Path):
        if isinstance(value, str):
            value = pathlib.Path(value)
        self._file_path = value

    @property
    def opacity(self) -> float:
        """The image's opacity, where 1 is fully opaque and 0 is invisible."""
        return self._opacity

    @opacity.setter
    def opacity(self, value: float):
        self._opacity = value

    @property
    def breakable_length(self) -> Unit:
        """The breakable length of the object.

        This is always ``ZERO``, meaning images cannot be broken across :obj:`.Flowable`
        lines.
        """
        return ZERO

    # Since Image isn't breakable (for now?), we only need to
    # implement complete rendering
    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        interface = ImageInterface(
            pos,
            None if flowable_line else self.parent.interface_for_children,
            self.scale,
            self.rotation,
            self.transform_origin,
            self.file_path,
            self.opacity,
        )
        interface.render()
        self.interfaces.append(interface)



from __future__ import annotations

from typing import Optional

from neoscore.core.directions import DirectionY
from neoscore.core.music_font import MusicFont
from neoscore.core.music_path import MusicPath
from neoscore.core.path_element import PathElement
from neoscore.core.pen import Pen
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO, Unit


class Stem(MusicPath):

    """A vertical note or chord stem path."""

    def __init__(
        self,
        start: PointDef,
        parent: PositionedObject,
        direction: DirectionY,
        height: Unit,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            start: Starting point for the stem
            parent: If no font is given, this or one of its ancestors must
                implement :obj:`.HasMusicFont`.
            direction: The direction the stem points
            height: The absolute-value height of the stem.
            font: If provided, this overrides any font found in the ancestor chain.
        """
        MusicPath.__init__(self, start, parent, font=font)
        self.pen = Pen(thickness=self.music_font.engraving_defaults["stemThickness"])

        self._direction = direction
        self._height = height
        # Draw stem path
        self.line_to(ZERO, self.height * self.direction.value)

    @property
    def height(self) -> Unit:
        """The absolute-value height of the stem."""
        return self._height

    @property
    def direction(self) -> DirectionY:
        """The direction the stem extends from the given position."""
        return self._direction

    @property
    def end_point(self) -> PathElement:
        """The outer point of the stem.

        In notes and chords this will typically be the line end which extends away from
        the notes.
        """
        return self.elements[1]
